import sempy.fabric as fabric
import time
from sempy_labs._helper_functions import resolve_dataset_id
from typing import Any, List, Optional, Union
from sempy._utils._log import log
import sempy_labs._icons as icons
from sempy_labs._helper_functions import resolve_workspace_name_and_id
from sempy.fabric.exceptions import FabricHTTPException


@log
def refresh_semantic_model(
    dataset: str,
    tables: Optional[Union[str, List[str]]] = None,
    partitions: Optional[Union[str, List[str]]] = None,
    refresh_type: str = "full",
    retry_count: int = 0,
    apply_refresh_policy: bool = True,
    max_parallelism: int = 10,
    workspace: Optional[str] = None,
):
    """
    Refreshes a semantic model.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    tables : str, List[str], default=None
        A string or a list of tables to refresh.
    partitions: str, List[str], default=None
        A string or a list of partitions to refresh. Partitions must be formatted as such: 'Table Name'[Partition Name].
    refresh_type : str, default="full"
        The type of processing to perform. Types align with the TMSL refresh command types: full, clearValues, calculate, dataOnly, automatic, and defragment. The add type isn't supported. Defaults to "full".
    retry_count : int, default=0
        Number of times the operation retries before failing.
    apply_refresh_policy : bool, default=True
        If an incremental refresh policy is defined, determines whether to apply the policy. Modes are true or false. If the policy isn't applied, the full process leaves partition definitions unchanged, and fully refreshes all partitions in the table. If commitMode is transactional, applyRefreshPolicy can be true or false. If commitMode is partialBatch, applyRefreshPolicy of true isn't supported, and applyRefreshPolicy must be set to false.
    max_parallelism : int, default=10
        Determines the maximum number of threads that can run the processing commands in parallel.
        This value aligns with the MaxParallelism property that can be set in the TMSL Sequence command or by using other methods.
        Defaults to 10.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    workspace = fabric.resolve_workspace_name(workspace)

    if isinstance(tables, str):
        tables = [tables]
    if isinstance(partitions, str):
        partitions = [partitions]

    objects: List[Any] = []

    if tables is not None:
        objects = objects + [{"table": table} for table in tables]
    if partitions is not None:

        def extract_names(partition):
            parts = partition.split("[")
            table_name = parts[0].strip("'")
            partition_name = parts[1].strip("]")
            return {"table": table_name, "partition": partition_name}

        objects = objects + [extract_names(partition) for partition in partitions]

    refresh_type = (
        refresh_type.lower().replace("only", "Only").replace("values", "Values")
    )

    if refresh_type not in icons.refreshTypes:
        raise ValueError(
            f"{icons.red_dot} Invalid refresh type. Refresh type must be one of these values: {icons.refreshTypes}."
        )

    if len(objects) == 0:
        requestID = fabric.refresh_dataset(
            dataset=dataset,
            workspace=workspace,
            refresh_type=refresh_type,
            retry_count=retry_count,
            apply_refresh_policy=apply_refresh_policy,
            max_parallelism=max_parallelism,
        )
    else:
        requestID = fabric.refresh_dataset(
            dataset=dataset,
            workspace=workspace,
            refresh_type=refresh_type,
            retry_count=retry_count,
            apply_refresh_policy=apply_refresh_policy,
            max_parallelism=max_parallelism,
            objects=objects,
        )
    print(
        f"{icons.in_progress} Refresh of the '{dataset}' semantic model within the '{workspace}' workspace is in progress..."
    )
    if len(objects) != 0:
        print(objects)

    while True:
        requestDetails = fabric.get_refresh_execution_details(
            dataset=dataset, refresh_request_id=requestID, workspace=workspace
        )
        status = requestDetails.status

        # Check if the refresh has completed
        if status == "Completed":
            break
        elif status == "Failed":
            raise ValueError(
                f"{icons.red_dot} The refresh of the '{dataset}' semantic model within the '{workspace}' workspace has failed."
            )
        elif status == "Cancelled":
            print(
                f"{icons.yellow_dot} The refresh of the '{dataset}' semantic model within the '{workspace}' workspace has been cancelled."
            )
            return

        time.sleep(3)

    print(
        f"{icons.green_dot} Refresh of the '{dataset}' semantic model within the '{workspace}' workspace is complete."
    )


@log
def cancel_dataset_refresh(
    dataset: str, request_id: Optional[str] = None, workspace: Optional[str] = None
):
    """
    Cancels the refresh of a semantic model which was executed via the `Enhanced Refresh API <https://learn.microsoft.com/power-bi/connect-data/asynchronous-refresh>`_

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    request_id : str, default=None
        The request id of a semantic model refresh.
        Defaults to finding the latest active refresh of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    rr = fabric.list_refresh_requests(dataset=dataset, workspace=workspace)
    rr_filt = rr[rr["Status"] == "Unknown"]

    if request_id is None:
        if len(rr_filt) == 0:
            raise ValueError(
                f"{icons.red_dot} There are no active Enhanced API refreshes of the '{dataset}' semantic model within the '{workspace}' workspace."
            )

        request_id = rr_filt["Request Id"].iloc[0]

    dataset_id = resolve_dataset_id(dataset=dataset, workspace=workspace)

    client = fabric.PowerBIRestClient()

    response = client.delete(
        f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/refreshes/{request_id}"
    )

    if response.status_code != 200:
        raise FabricHTTPException(response)
    print(
        f"{icons.green_dot} The '{request_id}' refresh request for the '{dataset}' semantic model within the '{workspace}' workspace has been cancelled."
    )



import sempy.fabric as fabric
from typing import Optional, List
from sempy._utils._log import log
import sempy_labs._icons as icons
from sempy.fabric.exceptions import FabricHTTPException
from sempy_labs._workspaces import assign_workspace_to_capacity
from sempy_labs.admin._basic_functions import (
    assign_workspaces_to_capacity,
    _list_capacities_meta,
    list_capacities,
)
from sempy_labs._helper_functions import (
    resolve_capacity_id,
    convert_to_alphanumeric_lowercase,
)
from sempy_labs._capacities import create_fabric_capacity


def migrate_settings(source_capacity: str, target_capacity: str):

    migrate_capacity_settings(
        source_capacity=source_capacity,
        target_capacity=target_capacity,
    )
    migrate_access_settings(
        source_capacity=source_capacity,
        target_capacity=target_capacity,
    )
    migrate_notification_settings(
        source_capacity=source_capacity,
        target_capacity=target_capacity,
    )
    migrate_spark_settings(
        source_capacity=source_capacity,
        target_capacity=target_capacity,
    )
    migrate_delegated_tenant_settings(
        source_capacity=source_capacity,
        target_capacity=target_capacity,
    )
    migrate_disaster_recovery_settings(
        source_capacity=source_capacity,
        target_capacity=target_capacity,
    )


@log
def migrate_workspaces(
    source_capacity: str,
    target_capacity: str,
    workspaces: Optional[str | List[str]] = None,
):
    """
    This function migrates the workspace(s) from one capacity to another capacity.
    Limitation: source & target capacities must be in the same region.
    If not all the workspaces succesfully migrated to the target capacity, the migrated workspaces will rollback to be assigned
    to the source capacity.

    Parameters
    ----------
    source_capacity : str
        Name of the source Fabric capacity.
    target_capacity : str
        Name of the target/destination Fabric capacity.
    workspaces : str | List[str], default=None
        The name of the workspace(s) specified will be reassigned from the source capacity to the target capacity.
        Defaults to None which will reassign all workspaces in the source capacity to the target capacity.
    """

    if isinstance(workspaces, str):
        workspaces = [workspaces]

    dfC = _list_capacities_meta()
    dfC_filt = dfC[dfC["Capacity Name"] == source_capacity]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} Invalid source capacity. The '{source_capacity}' capacity does not exist."
        )
    source_capacity_region = dfC_filt["Region"].iloc[0]
    source_capacity_id = dfC_filt["Capacity Id"].iloc[0]
    dfC_filt = dfC[dfC["Capacity Name"] == target_capacity]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} Invalid target capacity. The '{target_capacity}' capacity does not exist."
        )
    target_capacity_region = dfC_filt["Region"].iloc[0]
    target_capacity_state = dfC_filt["State"].iloc[0]

    if source_capacity_region != target_capacity_region:
        raise ValueError(
            f"{icons.red_dot} The '{source_capacity}' and '{target_capacity}' are not in the same region."
        )
    if target_capacity_state != "Active":
        raise ValueError(
            f"{icons.red_dot} The '{target_capacity}' target capacity is inactive. The capacity must be active in order for workspaces to be migrated."
        )

    dfW = fabric.list_workspaces(filter=f"capacityId eq '{source_capacity_id.upper()}'")
    if workspaces is None:
        workspace_count = len(dfW)
    else:
        workspace_count = len(workspaces)
    migrated_workspaces = []

    for i, r in dfW.iterrows():
        workspace = r["Name"]

        if workspaces is None or workspace in workspaces:
            pass
        else:
            continue

        if assign_workspace_to_capacity(
            capacity_name=target_capacity, workspace=workspace
        ):
            migrated_workspaces.append(workspace)

    if len(migrated_workspaces) < workspace_count:
        print(
            f"{icons.warning} Not all workspaces in the '{source_capacity}' capacity were migrated to the '{target_capacity}' capacity."
        )
        print(f"{icons.in_progress} Initiating rollback...")
        for i, r in dfW.iterrows():
            workspace = r["Name"]
            if workspace in migrated_workspaces:
                assign_workspace_to_capacity(
                    capacity_name=source_capacity, workspace=workspace
                )
        print(
            f"{icons.green_dot} Rollback of the workspaces to the '{source_capacity}' capacity is complete."
        )
    else:
        print(
            f"{icons.green_dot} All workspaces were migrated from the '{source_capacity}' capacity to the '{target_capacity}' capacity succesfully."
        )


@log
def migrate_capacities(
    azure_subscription_id: str,
    key_vault_uri: str,
    key_vault_tenant_id: str,
    key_vault_client_id: str,
    key_vault_client_secret: str,
    resource_group: str | dict,
    capacities: Optional[str | List[str]] = None,
    use_existing_rg_for_A_sku: bool = True,
    p_sku_only: bool = True,
):
    """
    This function creates new Fabric capacities for given A or P sku capacities and reassigns their workspaces to the newly created capacity.

    Parameters
    ----------
    azure_subscription_id : str
        The Azure subscription ID.
    key_vault_uri : str
        The name of the `Azure key vault <https://azure.microsoft.com/products/key-vault>`_ URI. Example: "https://<Key Vault Name>.vault.azure.net/"
    key_vault_tenant_id : str
        The name of the Azure key vault secret storing the Tenant ID.
    key_vault_client_id : str
        The name of the Azure key vault secret storing the Client ID.
    key_vault_client_secret : str
        The name of the Azure key vault secret storing the Client Secret.
    resource_group : str | dict
        The name of the Azure resource group.
        For A skus, this parameter will be ignored and the resource group used for the F sku will be the same as the A sku's resource group.
        For P skus, if this parameter is a string, it will use that resource group for all of the newly created F skus. If this parameter is a dictionary, it will use that mapping (capacity name -> resource group) for creating capacities with the mapped resource groups.
    capacities : str | List[str], default=None
        The capacity(ies) to migrate from A/P -> F sku.
        Defaults to None which migrates all accessible A/P sku capacities to F skus.
    use_existing_rg_for_A_sku : bool, default=True
        If True, the F sku inherits the resource group from the A sku (for A sku migrations)
    p_sku_only : bool, default=True
        If set to True, only migrates P skus. If set to False, migrates both P and A skus.
    """

    if isinstance(capacities, str):
        capacities = [capacities]

    p_sku_list = list(icons.sku_mapping.keys())

    dfC = list_capacities()

    if capacities is None:
        dfC_filt = dfC.copy()
    else:
        dfC_filt = dfC[dfC["Capacity Name"].isin(capacities)]

    if p_sku_only:
        dfC_filt = dfC_filt[dfC_filt["Sku"].str.startswith("P")]
    else:
        dfC_filt = dfC_filt[
            (dfC_filt["Sku"].str.startswith(("P", "A")))
            & (~dfC_filt["Sku"].str.startswith("PP"))
        ]

    dfC_filt = (
        dfC_filt.copy()
    )  # Something strange is happening here. Without this a key error on Display Name occurs

    if len(dfC_filt) == 0:
        print(f"{icons.info} There are no valid capacities to migrate.")
        return

    for _, r in dfC_filt.iterrows():
        cap_name = r["Capacity Name"]
        region = r["Region"]
        sku_size = r["Sku"]
        admins = r["Admins"]
        tgt_capacity = f"{convert_to_alphanumeric_lowercase(cap_name)}{icons.migrate_capacity_suffix}"

        # Check if target capacity exists
        dfC_tgt = dfC[dfC["Capacity Name"] == tgt_capacity]

        if sku_size[:1] == "A" and use_existing_rg_for_A_sku:
            rg = None
        else:
            if isinstance(resource_group, str):
                rg = resource_group
            elif isinstance(resource_group, dict):
                rg = resource_group.get(cap_name)
            else:
                raise ValueError(f"{icons.red_dot} Invalid 'resource_group' parameter.")

        if sku_size in p_sku_list:
            # Only create the capacity if it does not already exist
            if len(dfC_tgt) > 0:
                print(
                    f"{icons.info} Skipping creating a new capacity for '{cap_name}' as the '{tgt_capacity}' capacity already exists."
                )
            else:
                create_fabric_capacity(
                    capacity_name=tgt_capacity,
                    azure_subscription_id=azure_subscription_id,
                    key_vault_uri=key_vault_uri,
                    key_vault_tenant_id=key_vault_tenant_id,
                    key_vault_client_id=key_vault_client_id,
                    key_vault_client_secret=key_vault_client_secret,
                    resource_group=rg,
                    region=region,
                    sku=icons.sku_mapping.get(sku_size),
                    admin_members=admins,
                )
            # Migrate workspaces to new capacity
            assign_workspaces_to_capacity(
                source_capacity=cap_name, target_capacity=tgt_capacity, workspace=None
            )

            # Migrate settings to new capacity
            migrate_settings(source_capacity=cap_name, target_capacity=tgt_capacity)


@log
def migrate_capacity_settings(source_capacity: str, target_capacity: str):
    """
    This function migrates a capacity's settings to another capacity.

    Parameters
    ----------
    source_capacity : str
        Name of the source capacity.
    target_capacity : str
        Name of the target capacity.
    """

    dfC = list_capacities()
    dfC_filt = dfC[dfC["Capacity Name"] == source_capacity]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{source_capacity}' capacity does not exist."
        )
    source_capacity_id = dfC_filt["Capacity Id"].iloc[0].upper()
    dfC_filt = dfC[dfC["Capacity Name"] == target_capacity]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{target_capacity}' capacity does not exist."
        )
    target_capacity_id = dfC_filt["Capacity Id"].iloc[0].upper()

    workloads_params = "capacityCustomParameters?workloadIds=ADM&workloadIds=CDSA&workloadIds=DMS&workloadIds=RsRdlEngine&workloadIds=ScreenshotEngine&workloadIds=AS&workloadIds=QES&workloadIds=DMR&workloadIds=ESGLake&workloadIds=NLS&workloadIds=lake&workloadIds=TIPS&workloadIds=Kusto&workloadIds=Lakehouse&workloadIds=SparkCore&workloadIds=DI&workloadIds=Notebook&workloadIds=ML&workloadIds=ES&workloadIds=Reflex&workloadIds=Must&workloadIds=dmh&workloadIds=PowerBI&workloadIds=HLS"

    client = fabric.PowerBIRestClient()
    response_get_source = client.get(
        f"capacities/{source_capacity_id}/{workloads_params}"
    )
    if response_get_source.status_code != 200:
        raise FabricHTTPException(response_get_source)

    response_source_json = response_get_source.json().get(
        "capacityCustomParameters", {}
    )

    # Create payload for put request
    def remove_empty_workloads(data):
        keys_to_remove = [
            key for key, value in data.items() if not value["workloadCustomParameters"]
        ]
        for key in keys_to_remove:
            del data[key]

    remove_empty_workloads(response_source_json)

    settings_json = {}
    settings_json["capacityCustomParameters"] = {}

    for workload in response_source_json:
        if workload not in ["AI"]:
            settings_json["capacityCustomParameters"][workload] = {}
            settings_json["capacityCustomParameters"][workload][
                "workloadCustomParameters"
            ] = {}

            for workload_part in response_source_json[workload].values():
                for workload_item in workload_part:
                    setting_name = workload_item["name"]
                    setting_value = workload_item["value"]
                    if setting_value is None:
                        settings_json["capacityCustomParameters"][workload][
                            "workloadCustomParameters"
                        ][setting_name] = setting_value
                    elif isinstance(setting_value, bool):
                        settings_json["capacityCustomParameters"][workload][
                            "workloadCustomParameters"
                        ][setting_name] = bool(setting_value)
                    elif isinstance(setting_value, str):
                        settings_json["capacityCustomParameters"][workload][
                            "workloadCustomParameters"
                        ][setting_name] = str(setting_value)
                    else:
                        settings_json["capacityCustomParameters"][workload][
                            "workloadCustomParameters"
                        ][setting_name] = setting_value

    response_put = client.put(
        f"capacities/{target_capacity_id}/{workloads_params}",
        json=settings_json,
    )
    if response_put.status_code != 204:
        raise FabricHTTPException(response_put)

    print(
        f"{icons.green_dot} The capacity settings have been migrated from the '{source_capacity}' capacity to the '{target_capacity}' capacity."
    )


@log
def migrate_disaster_recovery_settings(source_capacity: str, target_capacity: str):
    """
    This function migrates a capacity's disaster recovery settings to another capacity.

    Parameters
    ----------
    source_capacity : str
        Name of the source capacity.
    target_capacity : str
        Name of the target capacity.
    """

    dfC = list_capacities()
    dfC_filt = dfC[dfC["Capacity Name"] == source_capacity]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{source_capacity}' capacity does not exist."
        )
    source_capacity_id = dfC_filt["Capacity Id"].iloc[0].upper()
    dfC_filt = dfC[dfC["Capacity Name"] == target_capacity]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{target_capacity}' capacity does not exist."
        )
    target_capacity_id = dfC_filt["Capacity Id"].iloc[0].upper()

    client = fabric.PowerBIRestClient()
    response_get_source = client.get(f"capacities/{source_capacity_id}/config")
    if response_get_source.status_code != 200:
        raise FabricHTTPException(response_get_source)

    request_body = {}
    value = response_get_source.json()["bcdr"]["config"]
    request_body["config"] = value

    response_put = client.put(
        f"capacities/{target_capacity_id}/fabricbcdr", json=request_body
    )

    if response_put.status_code != 202:
        raise FabricHTTPException(response_put)
    print(
        f"{icons.green_dot} The disaster recovery settings have been migrated from the '{source_capacity}' capacity to the '{target_capacity}' capacity."
    )


@log
def migrate_access_settings(source_capacity: str, target_capacity: str):
    """
    This function migrates the access settings from a source capacity to a target capacity.

    Parameters
    ----------
    source_capacity : str
        Name of the source capacity.
    target_capacity : str
        Name of the target capacity.
    """

    dfC = list_capacities()
    dfC_filt = dfC[dfC["Capacity Name"] == source_capacity]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{source_capacity}' capacity does not exist."
        )
    source_capacity_id = dfC_filt["Capacity Id"].iloc[0].upper()
    dfC_filt = dfC[dfC["Capacity Name"] == target_capacity]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{target_capacity}' capacity does not exist."
        )
    target_capacity_id = dfC_filt["Capacity Id"].iloc[0].upper()

    client = fabric.PowerBIRestClient()
    response_get_source = client.get(f"capacities/{source_capacity_id}")
    if response_get_source.status_code != 200:
        raise FabricHTTPException(response_get_source)

    access_settings = response_get_source.json().get("access", {})

    response_put = client.put(
        f"capacities/{target_capacity_id}/access",
        json=access_settings,
    )
    if response_put.status_code != 204:
        raise FabricHTTPException(response_put)

    print(
        f"{icons.green_dot} The access settings have been migrated from the '{source_capacity}' capacity to the '{target_capacity}' capacity."
    )


@log
def migrate_notification_settings(source_capacity: str, target_capacity: str):
    """
    This function migrates the notification settings from a source capacity to a target capacity.

    Parameters
    ----------
    source_capacity : str
        Name of the source capacity.
    target_capacity : str
        Name of the target capacity.
    """

    dfC = list_capacities()
    dfC_filt = dfC[dfC["Capacity Name"] == source_capacity]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{source_capacity}' capacity does not exist."
        )
    source_capacity_id = dfC_filt["Capacity Id"].iloc[0].upper()
    dfC_filt = dfC[dfC["Capacity Name"] == target_capacity]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{target_capacity}' capacity does not exist."
        )
    target_capacity_id = dfC_filt["Capacity Id"].iloc[0].upper()

    client = fabric.PowerBIRestClient()
    response_get_source = client.get(f"capacities/{source_capacity_id}")
    if response_get_source.status_code != 200:
        raise FabricHTTPException(response_get_source)

    notification_settings = response_get_source.json().get(
        "capacityNotificationSettings", {}
    )

    response_put = client.put(
        f"capacities/{target_capacity_id}/notificationSettings",
        json=notification_settings,
    )
    if response_put.status_code != 204:
        raise FabricHTTPException(response_put)

    print(
        f"{icons.green_dot} The notification settings have been migrated from the '{source_capacity}' capacity to the '{target_capacity}' capacity."
    )


@log
def migrate_delegated_tenant_settings(source_capacity: str, target_capacity: str):
    """
    This function migrates the delegated tenant settings from a source capacity to a target capacity.

    Parameters
    ----------
    source_capacity : str
        Name of the source capacity.
    target_capacity : str
        Name of the target capacity.
    """

    dfC = list_capacities()

    dfC_filt = dfC[dfC["Capacity Name"] == source_capacity]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{source_capacity}' capacity does not exist."
        )
    source_capacity_id = dfC_filt["Capacity Id"].iloc[0].upper()

    dfC_filt = dfC[dfC["Capacity Name"] == target_capacity]
    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{target_capacity}' capacity does not exist."
        )
    target_capacity_id = dfC_filt["Capacity Id"].iloc[0].upper()

    client = fabric.FabricRestClient()
    response_get = client.get("v1/admin/capacities/delegatedTenantSettingOverrides")

    if response_get.status_code != 200:
        raise FabricHTTPException(response_get)

    response_json = response_get.json().get("Overrides", [])

    for o in response_json:
        if o.get("id").upper() == source_capacity_id:
            for setting in o.get("tenantSettings", []):
                setting_name = setting.get("settingName")
                feature_switch = {
                    "switchId": -1,
                    "switchName": setting_name,
                    "isEnabled": setting.get("enabled", False),
                    "isGranular": setting.get("canSpecifySecurityGroups", False),
                    "allowedSecurityGroups": [
                        {
                            "id": group.get("graphId"),
                            "name": group.get("name"),
                            "isEmailEnabled": False,
                        }
                        for group in setting.get("enabledSecurityGroups", [])
                    ],
                    "deniedSecurityGroups": [
                        {
                            "id": group.get("graphId"),
                            "name": group.get("name"),
                            "isEmailEnabled": False,
                        }
                        for group in setting.get("excludedSecurityGroups", [])
                    ],
                }

                payload = {"featureSwitches": [feature_switch], "properties": []}

                client = fabric.PowerBIRestClient()
                response_put = client.put(
                    f"metadata/tenantsettings/selfserve?capacityObjectId={target_capacity_id}",
                    json=payload,
                )
                if response_put.status_code != 200:
                    raise FabricHTTPException(response_put)

                print(
                    f"{icons.green_dot} The delegated tenant settings for the '{setting_name}' feature switch of the '{source_capacity}' capacity have been migrated to the '{target_capacity}' capacity."
                )


@log
def migrate_spark_settings(source_capacity: str, target_capacity: str):
    """
    This function migrates a capacity's spark settings to another capacity.

    Requirement: The target capacity must be able to accomodate the spark pools being migrated from the source capacity.

    Parameters
    ----------
    source_capacity : str
        Name of the source capacity.
    target_capacity : str
        Name of the target capacity.
    """

    source_capacity_id = resolve_capacity_id(capacity_name=source_capacity)
    target_capacity_id = resolve_capacity_id(capacity_name=target_capacity)
    client = fabric.PowerBIRestClient()

    # Get source capacity server dns
    response = client.get(f"metadata/capacityInformation/{source_capacity_id}")
    if response.status_code != 200:
        raise FabricHTTPException(response)

    source_server_dns = response.json().get("capacityDns")
    source_url = f"{source_server_dns}/webapi/capacities"

    # Get target capacity server dns
    response = client.get(f"metadata/capacityInformation/{target_capacity_id}")
    if response.status_code != 200:
        raise FabricHTTPException(response)

    target_server_dns = response.json().get("capacityDns")
    target_url = f"{target_server_dns}/webapi/capacities"

    # Construct get and put URLs
    end_url = "workloads/SparkCore/SparkCoreService/automatic/v1/sparksettings"
    get_url = f"{source_url}/{source_capacity_id}/{end_url}"
    put_url = f"{target_url}/{target_capacity_id}/{end_url}/content"

    # Get source capacity spark settings
    response = client.get(get_url)
    if response.status_code != 200:
        raise FabricHTTPException(response)

    payload = response.json().get("content")

    # Update target capacity spark settings
    response_put = client.put(put_url, json=payload)

    if response_put.status_code != 200:
        raise FabricHTTPException(response_put)
    print(
        f"{icons.green_dot} The spark settings have been migrated from the '{source_capacity}' capacity to the '{target_capacity}' capacity."
    )


@log
def migrate_fabric_trial_capacity(
    azure_subscription_id: str,
    key_vault_uri: str,
    key_vault_tenant_id: str,
    key_vault_client_id: str,
    key_vault_client_secret: str,
    resource_group: str,
    source_capacity: str,
    target_capacity: str,
    target_capacity_sku: str = "F64",
    target_capacity_region: Optional[str] = None,
    target_capacity_admin_members: Optional[str | List[str]] = None,
):
    """
    This function migrates a Fabric trial capacity to a Fabric capacity. If the 'target_capacity' does not exist, it is created with the relevant target capacity parameters (sku, region, admin members).

    Parameters
    ----------
    azure_subscription_id : str
        The Azure subscription ID.
    key_vault_uri : str
        The name of the `Azure key vault <https://azure.microsoft.com/products/key-vault>`_ URI. Example: "https://<Key Vault Name>.vault.azure.net/"
    key_vault_tenant_id : str
        The name of the Azure key vault secret storing the Tenant ID.
    key_vault_client_id : str
        The name of the Azure key vault secret storing the Client ID.
    key_vault_client_secret : str
        The name of the Azure key vault secret storing the Client Secret.
    resource_group : str
        The name of the Azure resource group.
    source_capacity : str
        The name of the Fabric trial capacity.
    target_capacity : str
        The name of the new Fabric capacity (F SKU). If this capacity does not exist, it will be created.
    target_capacity_sku : str, default="F64"
        If the target capacity does not exist, this property sets the SKU size for the target capacity.
    target_capacity_region : str, default=None
        If the target capacity does not exist, this property sets the region for the target capacity.
        Defaults to None which resolves to the region in which the Trial SKU exists.
    target_capacity_admin_members : str, default=None
        If the target capacity does not exist, this property sets the admin members for the target capacity.
        Defaults to None which resolves to the admin members on the Trial SKU.
    """

    notebook_workspace_id = fabric.get_notebook_workspace_id()
    dfW = fabric.list_workspaces(filter=f"id eq '{notebook_workspace_id}'")
    notebook_capacity_id = dfW["Capacity Id"].iloc[0].lower()

    dfC = list_capacities()
    dfC_filt = dfC[dfC["Capacity Name"] == source_capacity]

    if len(dfC_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The {source_capacity}' capacity does not exist."
        )

    source_capacity_sku = dfC_filt["Sku"].iloc[0]
    if not source_capacity_sku.startswith("FT"):
        raise ValueError(
            f"{icons.red_dot} This function is for migrating Fabric trial capacites to Fabric capacities."
        )

    source_capacity_id = dfC_filt["Capacity Id"].iloc[0].lower()
    if source_capacity_id == notebook_capacity_id:
        print(
            f"{icons.warning} The '{source_capacity}' capacity cannot be both the source capacity as well as the capacity in which the notebook is running."
        )
        return

    # Use same region as source capacity if no region is specified
    if target_capacity_region is None:
        target_capacity_region = dfC_filt["Region"].iloc[0]

    # Use same admins as source capacity
    if isinstance(target_capacity_admin_members, str):
        target_capacity_admin_members = [target_capacity_admin_members]

    if target_capacity_admin_members is None:
        target_capacity_admin_members = dfC_filt["Admins"].iloc[0]

    dfC_filt = dfC[dfC["Capacity Name"] == target_capacity]
    if len(dfC_filt) == 0:
        create_fabric_capacity(
            capacity_name=target_capacity,
            azure_subscription_id=azure_subscription_id,
            key_vault_uri=key_vault_uri,
            key_vault_tenant_id=key_vault_tenant_id,
            key_vault_client_id=key_vault_client_id,
            key_vault_client_secret=key_vault_client_secret,
            resource_group=resource_group,
            region=target_capacity_region,
            admin_members=target_capacity_admin_members,
            sku=target_capacity_sku,
        )

    assign_workspaces_to_capacity(
        source_capacity=source_capacity,
        target_capacity=target_capacity,
        workspace=None,
    )

    # This migrates all the capacity settings
    migrate_settings(
        source_capacity=source_capacity,
        target_capacity=target_capacity,
    )



import sempy.fabric as fabric
import pandas as pd
import sempy_labs._icons as icons
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    lro,
    pagination,
    _decode_b64,
)
from sempy.fabric.exceptions import FabricHTTPException


def list_data_pipelines(workspace: Optional[str] = None) -> pd.DataFrame:
    """
    Shows the data pipelines within a workspace.

    Parameters
    ----------
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the data pipelines within a workspace.
    """

    df = pd.DataFrame(columns=["Data Pipeline Name", "Data Pipeline ID", "Description"])

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    client = fabric.FabricRestClient()
    response = client.get(f"/v1/workspaces/{workspace_id}/dataPipelines")
    if response.status_code != 200:
        raise FabricHTTPException(response)

    responses = pagination(client, response)

    for r in responses:
        for v in r.get("value", []):
            new_data = {
                "Data Pipeline Name": v.get("displayName"),
                "Data Pipeline ID": v.get("id"),
                "Description": v.get("description"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def create_data_pipeline(
    name: str, description: Optional[str] = None, workspace: Optional[str] = None
):
    """
    Creates a Fabric data pipeline.

    Parameters
    ----------
    name: str
        Name of the data pipeline.
    description : str, default=None
        A description of the environment.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    request_body = {"displayName": name}

    if description:
        request_body["description"] = description

    client = fabric.FabricRestClient()
    response = client.post(
        f"/v1/workspaces/{workspace_id}/dataPipelines", json=request_body
    )

    lro(client, response, status_codes=[201, 202])

    print(
        f"{icons.green_dot} The '{name}' data pipeline has been created within the '{workspace}' workspace."
    )


def delete_data_pipeline(name: str, workspace: Optional[str] = None):
    """
    Deletes a Fabric data pipeline.

    Parameters
    ----------
    name: str
        Name of the data pipeline.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    item_id = fabric.resolve_item_id(
        item_name=name, type="DataPipeline", workspace=workspace
    )

    client = fabric.FabricRestClient()
    response = client.delete(f"/v1/workspaces/{workspace_id}/dataPipelines/{item_id}")

    if response.status_code != 200:
        raise FabricHTTPException(response)

    print(
        f"{icons.green_dot} The '{name}' data pipeline within the '{workspace}' workspace has been deleted."
    )


def get_data_pipeline_definition(
    name: str, workspace: Optional[str] = None, decode: bool = True
) -> dict | pd.DataFrame:
    """
    Obtains the definition of a data pipeline.

    Parameters
    ----------
    name : str
        The name of the data pipeline.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    decode : bool, default=True
        decode : bool, default=True
        If True, decodes the data pipeline definition file into .json format.
        If False, obtains the data pipeline definition file a pandas DataFrame format.

    Returns
    -------
    dict | pandas.DataFrame
        A pandas dataframe showing the data pipelines within a workspace.
    """

    workspace = fabric.resolve_workspace_name(workspace)
    workspace_id = fabric.resolve_workspace_id(workspace)
    item_id = fabric.resolve_item_id(
        item_name=name, type="DataPipeline", workspace=workspace
    )

    client = fabric.FabricRestClient()
    response = client.post(
        f"/v1/workspaces/{workspace_id}/dataPipelines/{item_id}/getDefinition"
    )
    result = lro(client, response).json()

    df = pd.json_normalize(result["definition"]["parts"])

    if not decode:
        return df
    content = df[df["path"] == "pipeline-content.json"]
    payload = content["payload"].iloc[0]

    return _decode_b64(payload)



import sempy.fabric as fabric
from typing import Optional, List, Union
from uuid import UUID
import sempy_labs._icons as icons
from sempy.fabric.exceptions import FabricHTTPException
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    pagination,
)
import numpy as np
import pandas as pd
import time


def list_workspaces(
    top: Optional[int] = 5000, skip: Optional[int] = None
) -> pd.DataFrame:
    """
    Lists workspaces for the organization. This function is the admin version of list_workspaces.

    Parameters
    ----------
    top : int, default=5000
        Returns only the first n results. This parameter is mandatory and must be in the range of 1-5000.
    skip : int, default=None
        Skips the first n results. Use with top to fetch results beyond the first 5000.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of workspaces for the organization.
    """

    df = pd.DataFrame(
        columns=[
            "Id",
            "Is Read Only",
            "Is On Dedicated Capacity",
            "Type",
            "Name",
            "Capacity Id",
            "Default Dataset Storage Format",
            "Pipeline Id",
            "Has Workspace Level Settings",
        ]
    )

    url = f"/v1.0/myorg/admin/groups?$top={top}"
    if skip is not None:
        url = f"{url}&$skip={skip}"

    client = fabric.PowerBIRestClient()
    response = client.get(url)

    if response.status_code != 200:
        raise FabricHTTPException(response)

    for v in response.json().get("value", []):
        capacity_id = v.get("capacityId")
        if capacity_id:
            capacity_id = capacity_id.lower()
        new_data = {
            "Id": v.get("id"),
            "Is Read Only": v.get("isReadOnly"),
            "Is On Dedicated Capacity": v.get("isOnDedicatedCapacity"),
            "Capacity Id": capacity_id,
            "Default Dataset Storage Format": v.get("defaultDatasetStorageFormat"),
            "Type": v.get("type"),
            "Name": v.get("name"),
            "State": v.get("state"),
            "Pipeline Id": v.get("pipelineId"),
            "Has Workspace Level Settings": v.get("hasWorkspaceLevelSettings"),
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    bool_cols = [
        "Is Read Only",
        "Is On Dedicated Capacity",
        "Has Workspace Level Settings",
    ]
    df[bool_cols] = df[bool_cols].astype(bool)

    return df


def assign_workspaces_to_capacity(
    source_capacity: str,
    target_capacity: str,
    workspace: Optional[str | List[str]] = None,
):
    """
    Assigns a workspace to a capacity. This function is the admin version.

    Parameters
    ----------
    source_capacity : str
        The name of the source capacity.
    target_capacity : str
        The name of the target capacity.
    workspace : str | List[str], default=None
        The name of the workspace(s).
        Defaults to None which resolves to migrating all workspaces within the source capacity to the target capacity.
    """

    if isinstance(workspace, str):
        workspace = [workspace]

    dfC = list_capacities()
    dfC_filt = dfC[dfC["Capacity Name"] == source_capacity]
    source_capacity_id = dfC_filt["Capacity Id"].iloc[0]

    dfC_filt = dfC[dfC["Capacity Name"] == target_capacity]
    target_capacity_id = dfC_filt["Capacity Id"].iloc[0]

    if workspace is None:
        # workspaces = fabric.list_workspaces(
        #    filter=f"capacityId eq '{source_capacity_id.upper()}'"
        # )["Id"].values
        dfW = list_workspaces()
        dfW = dfW[dfW["Capacity Id"].str.upper() == source_capacity_id.upper()]
        workspaces = dfW["Id"].tolist()
    else:
        dfW = list_workspaces()
        workspaces = dfW[dfW["Name"].isin(workspace)]["Id"].tolist()

    workspaces = np.array(workspaces)
    batch_size = 999
    for i in range(0, len(workspaces), batch_size):
        batch = workspaces[i : i + batch_size].tolist()
        request_body = {
            "capacityMigrationAssignments": [
                {
                    "targetCapacityObjectId": target_capacity_id.upper(),
                    "workspacesToAssign": batch,
                }
            ]
        }

        client = fabric.PowerBIRestClient()
        response = client.post(
            "/v1.0/myorg/admin/capacities/AssignWorkspaces",
            json=request_body,
        )

        if response.status_code != 200:
            raise FabricHTTPException(response)
    print(
        f"{icons.green_dot} The workspaces have been assigned to the '{target_capacity}' capacity."
    )


def list_capacities() -> pd.DataFrame:
    """
    Shows the a list of capacities and their properties. This function is the admin version.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the capacities and their properties
    """

    df = pd.DataFrame(
        columns=["Capacity Id", "Capacity Name", "Sku", "Region", "State", "Admins"]
    )

    client = fabric.PowerBIRestClient()
    response = client.get("/v1.0/myorg/admin/capacities")
    if response.status_code != 200:
        raise FabricHTTPException(response)

    responses = pagination(client, response)

    for r in responses:
        for i in r.get("value", []):
            new_data = {
                "Capacity Id": i.get("id").lower(),
                "Capacity Name": i.get("displayName"),
                "Sku": i.get("sku"),
                "Region": i.get("region"),
                "State": i.get("state"),
                "Admins": [i.get("admins", [])],
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def list_tenant_settings() -> pd.DataFrame:
    """
    Lists all tenant settings.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the tenant settings.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/tenants/list-tenant-settings?tabs=HTTP

    client = fabric.FabricRestClient()
    response = client.get("/v1/admin/tenantsettings")

    if response.status_code != 200:
        raise FabricHTTPException(response)

    df = pd.DataFrame(
        columns=[
            "Setting Name",
            "Title",
            "Enabled",
            "Can Specify Security Groups",
            "Tenant Setting Group",
            "Enabled Security Groups",
        ]
    )

    for i in response.json().get("tenantSettings", []):
        new_data = {
            "Setting Name": i.get("settingName"),
            "Title": i.get("title"),
            "Enabled": i.get("enabled"),
            "Can Specify Security Groups": i.get("canSpecifySecurityGroups"),
            "Tenant Setting Group": i.get("tenantSettingGroup"),
            "Enabled Security Groups": [i.get("enabledSecurityGroups", [])],
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    bool_cols = ["Enabled", "Can Specify Security Groups"]
    df[bool_cols] = df[bool_cols].astype(bool)

    return df


def _list_capacities_meta() -> pd.DataFrame:

    df = pd.DataFrame(
        columns=["Capacity Id", "Capacity Name", "Sku", "Region", "State", "Admins"]
    )

    client = fabric.PowerBIRestClient()
    try:
        response = client.get("/v1.0/myorg/admin/capacities")
    except Exception as e:
        if e.status_code not in [200, 401]:
            raise FabricHTTPException(response)
        elif e.status_code == 401:
            response = client.get("/v1.0/myorg/capacities")

    for i in response.json().get("value", []):
        new_data = {
            "Capacity Id": i.get("id").lower(),
            "Capacity Name": i.get("displayName"),
            "Sku": i.get("sku"),
            "Region": i.get("region"),
            "State": i.get("state"),
            "Admins": [i.get("admins", [])],
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def unassign_workspaces_from_capacity(workspaces: str | List[str]):
    """
    Unassigns workspace(s) from their capacity. This function is the admin version of list_workspaces.

    Parameters
    ----------
    workspaces : str | List[str]
        The Fabric workspace name(s).
    """

    # https://learn.microsoft.com/en-us/rest/api/power-bi/admin/capacities-unassign-workspaces-from-capacity

    if isinstance(workspaces, str):
        workspaces = [workspaces]

    payload = {"workspacesToUnassign": workspaces}

    client = fabric.PowerBIRestClient()
    response = client.post(
        "/v1.0/myorg/admin/capacities/UnassignWorkspaces",
        json=payload,
    )

    if response.status_code != 200:
        raise FabricHTTPException(response)

    print(f"{icons.green_dot} The workspaces have been unassigned.")


def list_external_data_shares():
    """
    Lists external data shares in the tenant. This function is for admins.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of external data shares in the tenant.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/external-data-shares/list-external-data-shares?tabs=HTTP

    df = pd.DataFrame(
        columns=[
            "External Data Share Id",
            "Paths",
            "Creater Principal Id",
            "Creater Principal Name",
            "Creater Principal Type",
            "Creater Principal UPN",
            "Recipient UPN",
            "Status",
            "Expiration Time UTC",
            "Workspace Id",
            "Item Id",
            "Invitation URL",
        ]
    )

    client = fabric.FabricRestClient()
    response = client.get("/v1/admin/items/externalDataShares")

    if response.status_code != 200:
        raise FabricHTTPException(response)

    for i in response.json().get("value", []):
        cp = i.get("creatorPrincipal", {})
        new_data = {
            "External Data Share Id": i.get("id"),
            "Paths": [i.get("paths", [])],
            "Creater Principal Id": cp.get("id"),
            "Creater Principal Name": cp.get("displayName"),
            "Creater Principal Type": cp.get("type"),
            "Creater Principal UPN": cp.get("userDetails", {}).get("userPrincipalName"),
            "Recipient UPN": i.get("recipient", {}).get("userPrincipalName"),
            "Status": i.get("status"),
            "Expiration Time UTC": i.get("expirationTimeUtc"),
            "Workspace Id": i.get("workspaceId"),
            "Item Id": i.get("itemId"),
            "Invitation URL": i.get("invitationUrl"),
        }

        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    date_time_columns = ["Expiration Time UTC"]
    df[date_time_columns] = pd.to_datetime(df[date_time_columns])

    return df


def revoke_external_data_share(
    external_data_share_id: UUID, item_id: UUID, workspace: str
):
    """
    Revokes the specified external data share. Note: This action cannot be undone.

    Parameters
    ----------
    external_data_share_id : UUID
        The external data share ID.
    item_id : int, default=None
        The Item ID
    workspace : str
        The Fabric workspace name.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/external-data-shares/revoke-external-data-share?tabs=HTTP

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    client = fabric.FabricRestClient()
    response = client.post(
        f"/v1/admin/workspaces/{workspace_id}/items/{item_id}/externalDataShares/{external_data_share_id}/revoke"
    )

    if response.status_code != 200:
        raise FabricHTTPException(response)

    print(
        f"{icons.green_dot} The '{external_data_share_id}' external data share for the '{item_id}' item within the '{workspace}' workspace has been revoked."
    )


def list_capacities_delegated_tenant_settings(
    return_dataframe: bool = True,
) -> Optional[pd.DataFrame | dict]:
    """
    Returns list of tenant setting overrides that override at the capacities.

    Parameters
    ----------
    return_dataframe : bool, default=True
        If True, returns a dataframe. If False, returns a dictionary

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of tenant setting overrides that override at the capacities.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/tenants/list-capacities-tenant-settings-overrides?tabs=HTTP

    df = pd.DataFrame(
        columns=[
            "Capacity Id",
            "Setting Name",
            "Setting Title",
            "Setting Enabled",
            "Can Specify Security Groups",
            "Enabled Security Groups",
            "Tenant Setting Group",
            "Tenant Setting Properties",
            "Delegate to Workspace",
            "Delegated From",
        ]
    )

    client = fabric.FabricRestClient()
    response = client.get("/v1/admin/capacities/delegatedTenantSettingOverrides")

    if response.status_code != 200:
        raise FabricHTTPException(response)

    responses = pagination(client, response)

    if return_dataframe:
        for r in responses:
            for i in r.get("Overrides", []):
                tenant_settings = i.get("tenantSettings", [])
                for setting in tenant_settings:
                    new_data = {
                        "Capacity Id": i.get("id"),
                        "Setting Name": setting.get("settingName"),
                        "Setting Title": setting.get("title"),
                        "Setting Enabled": setting.get("enabled"),
                        "Can Specify Security Groups": setting.get(
                            "canSpecifySecurityGroups"
                        ),
                        "Enabled Security Groups": [
                            setting.get("enabledSecurityGroups", [])
                        ],
                        "Tenant Setting Group": setting.get("tenantSettingGroup"),
                        "Tenant Setting Properties": [setting.get("properties", [])],
                        "Delegate to Workspace": setting.get("delegateToWorkspace"),
                        "Delegated From": setting.get("delegatedFrom"),
                    }

                    df = pd.concat(
                        [df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )

            bool_cols = [
                "Enabled Security Groups",
                "Can Specify Security Groups",
                "Delegate to Workspace",
            ]
            df[bool_cols] = df[bool_cols].astype(bool)

            return df
    else:
        combined_response = {
            "overrides": [],
            "continuationUri": "",
            "continuationToken": "",
        }
        for r in responses:
            combined_response["overrides"].extend(r["overrides"])
            combined_response["continuationUri"] = r["continuationUri"]
            combined_response["continuationToken"] = r["continuationToken"]

        return combined_response


def scan_workspaces(
    data_source_details: bool = False,
    dataset_schema: bool = False,
    dataset_expressions: bool = False,
    lineage: bool = False,
    artifact_users: bool = False,
    workspace: Optional[str | List[str]] = None,
) -> dict:

    if workspace is None:
        workspace = fabric.resolve_workspace_name()

    if isinstance(workspace, str):
        workspace = [workspace]

    workspace_list = []

    for w in workspace:
        workspace_list.append(fabric.resolve_workspace_id(w))

    client = fabric.PowerBIRestClient()
    request_body = {"workspaces": workspace_list}

    response_clause = f"/v1.0/myorg/admin/workspaces/getInfo?lineage={lineage}&datasourceDetails={data_source_details}&datasetSchema={dataset_schema}&datasetExpressions={dataset_expressions}&getArtifactUsers={artifact_users}"
    response = client.post(response_clause, json=request_body)

    if response.status_code != 202:
        raise FabricHTTPException(response)
    scan_id = response.json()["id"]
    scan_status = response.json().get("status")
    while scan_status not in ["Succeeded", "Failed"]:
        time.sleep(1)
        response = client.get(f"/v1.0/myorg/admin/workspaces/scanStatus/{scan_id}")
        scan_status = response.json().get("status")
    if scan_status == "Failed":
        raise FabricHTTPException(response)
    response = client.get(f"/v1.0/myorg/admin/workspaces/scanResult/{scan_id}")
    if response.status_code != 200:
        raise FabricHTTPException(response)

    return response.json()


def list_datasets() -> pd.DataFrame:
    """
    Shows a list of datasets for the organization.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of datasets for the organization.
    """

    # https://learn.microsoft.com/en-us/rest/api/power-bi/admin/datasets-get-datasets-as-admin

    df = pd.DataFrame(
        columns=[
            "Dataset Id",
            "Dataset Name",
            "Web URL",
            "Add Rows API Enabled",
            "Configured By",
            "Is Refreshable",
            "Is Effective Identity Required",
            "Is Effective Identity Roles Required",
            "Target Storage Mode",
            "Created Date",
            "Content Provider Type",
            "Create Report Embed URL",
            "QnA Embed URL",
            "Upstream Datasets",
            "Users",
            "Is In Place Sharing Enabled",
            "Workspace Id",
            "Auto Sync Read Only Replicas",
            "Max Read Only Replicas",
        ]
    )

    client = fabric.PowerBIRestClient()
    response = client.get("/v1.0/myorg/admin/datasets")

    if response.status_code != 200:
        raise FabricHTTPException(response)

    for v in response.json().get("value", []):
        new_data = {
            "Dataset Id": v.get("id"),
            "Dataset Name": v.get("name"),
            "Web URL": v.get("webUrl"),
            "Add Rows API Enabled": v.get("addRowsAPIEnabled"),
            "Configured By": v.get("configuredBy"),
            "Is Refreshable": v.get("isRefreshable"),
            "Is Effective Identity Required": v.get("isEffectiveIdentityRequired"),
            "Is Effective Identity Roles Required": v.get(
                "isEffectiveIdentityRolesRequired"
            ),
            "Target Storage Mode": v.get("targetStorageMode"),
            "Created Date": pd.to_datetime(v.get("createdDate")),
            "Content Provider Type": v.get("contentProviderType"),
            "Create Report Embed URL": v.get("createReportEmbedURL"),
            "QnA Embed URL": v.get("qnaEmbedURL"),
            "Upstream Datasets": v.get("upstreamDatasets", []),
            "Users": v.get("users", []),
            "Is In Place Sharing Enabled": v.get("isInPlaceSharingEnabled"),
            "Workspace Id": v.get("workspaceId"),
            "Auto Sync Read Only Replicas": v.get("queryScaleOutSettings", {}).get(
                "autoSyncReadOnlyReplicas"
            ),
            "Max Read Only Replicas": v.get("queryScaleOutSettings", {}).get(
                "maxReadOnlyReplicas"
            ),
        }
        df = pd.concat([df, pd.DataFrame([new_data])], ignore_index=True)

    bool_cols = [
        "Add Rows API Enabled",
        "Is Refreshable",
        "Is Effective Identity Required",
        "Is Effective Identity Roles Required",
        "Is In Place Sharing Enabled",
        "Auto Sync Read Only Replicas",
    ]
    df[bool_cols] = df[bool_cols].astype(bool)

    df["Created Date"] = pd.to_datetime(df["Created Date"])
    df["Max Read Only Replicas"] = df["Max Read Only Replicas"].astype(int)

    return df


def list_item_access_details(
    item_name: str, type: str, workspace: Optional[str] = None
) -> pd.DataFrame:
    """
    Returns a list of users (including groups and service principals) and lists their workspace roles.

    Parameters
    ----------
    item_name : str
        Name of the Fabric item.
    type : str
        Type of Fabric item.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of users (including groups and service principals) and lists their workspace roles.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/items/list-item-access-details?tabs=HTTP

    workspace = fabric.resolve_workspace_name(workspace)
    workspace_id = fabric.resolve_workspace_id(workspace)
    item_id = fabric.resolve_item_id(
        item_name=item_name, type=type, workspace=workspace
    )

    df = pd.DataFrame(
        columns=[
            "User Id",
            "User Name",
            "User Type",
            "User Principal Name",
            "Item Name",
            "Item Type",
            "Item Id",
            "Permissions",
            "Additional Permissions",
        ]
    )
    client = fabric.FabricRestClient()
    response = client.get(f"/v1/admin/workspaces/{workspace_id}/items/{item_id}/users")

    if response.status_code != 200:
        raise FabricHTTPException(response)

    for v in response.json().get("accessDetails", []):
        new_data = {
            "User Id": v.get("principal", {}).get("id"),
            "User Name": v.get("principal", {}).get("displayName"),
            "User Type": v.get("principal", {}).get("type"),
            "User Principal Name": v.get("principal", {})
            .get("userDetails", {})
            .get("userPrincipalName"),
            "Item Type": v.get("itemAccessDetails", {}).get("type"),
            "Permissions": v.get("itemAccessDetails", {}).get("permissions"),
            "Additional Permissions": v.get("itemAccessDetails", {}).get(
                "additionalPermissions"
            ),
            "Item Name": item_name,
            "Item Id": item_id,
        }
        df = pd.concat([df, pd.DataFrame([new_data])], ignore_index=True)

    return df


def list_access_entities(
    user_email_address: str,
) -> pd.DataFrame:
    """
    Shows a list of permission details for Fabric and PowerBI items the specified user can access.

    Parameters
    ----------
    user_email_address : str
        The user's email address.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of permission details for Fabric and PowerBI items the specified user can access.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/users/list-access-entities?tabs=HTTP

    df = pd.DataFrame(
        columns=[
            "Item Id",
            "Item Name",
            "Item Type",
            "Permissions",
            "Additional Permissions",
        ]
    )
    client = fabric.FabricRestClient()
    response = client.get(f"/v1/admin/users/{user_email_address}/access")

    if response.status_code != 200:
        raise FabricHTTPException(response)

    responses = pagination(client, response)

    for r in responses:
        for v in r.get("accessEntities", []):
            new_data = {
                "Item Id": v.get("id"),
                "Item Name": v.get("displayName"),
                "Item Type": v.get("itemAccessDetails", {}).get("type"),
                "Permissions": v.get("itemAccessDetails", {}).get("permissions"),
                "Additional Permissions": v.get("itemAccessDetails", {}).get(
                    "additionalPermissions"
                ),
            }
            df = pd.concat([df, pd.DataFrame([new_data])], ignore_index=True)

    return df


def list_workspace_access_details(
    workspace: Optional[Union[str, UUID]] = None
) -> pd.DataFrame:
    """
    Shows a list of users (including groups and Service Principals) that have access to the specified workspace.

    Parameters
    ----------
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of users (including groups and Service Principals) that have access to the specified workspace.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/workspaces/list-workspace-access-details?tabs=HTTP

    workspace_name = fabric.resolve_workspace_name(workspace)
    workspace_id = fabric.resolve_workspace_id(workspace_name)

    df = pd.DataFrame(
        columns=[
            "User Id",
            "User Name",
            "User Type",
            "Workspace Name",
            "Workspace Id",
            "Workspace Role",
        ]
    )
    client = fabric.FabricRestClient()
    response = client.get(f"/v1/admin/workspaces/{workspace_id}/users")
    if response.status_code != 200:
        raise FabricHTTPException(response)

    for v in response.json().get("accessDetails", []):
        new_data = {
            "User Id": v.get("principal", {}).get("id"),
            "User Name": v.get("principal", {}).get("displayName"),
            "User Type": v.get("principal", {}).get("type"),
            "Workspace Name": workspace_name,
            "Workspace Id": workspace_id,
            "Workspace Role": v.get("workspaceAccessDetails", {}).get("workspaceRole"),
        }
        df = pd.concat([df, pd.DataFrame([new_data])], ignore_index=True)

    return df


def list_items(
    capacity_name: Optional[str] = None,
    workspace: Optional[str] = None,
    state: Optional[str] = None,
    type: Optional[str] = None,
) -> pd.DataFrame:
    """
    Shows a list of active Fabric and PowerBI items.

    Parameters
    ----------
    capacity_name : str, default=None
        The capacity name.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    state : str, default=None
        The item state.
    type : str, default=None
        The item type.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of active Fabric and Power BI items.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/items/list-items?tabs=HTTP

    url = "/v1/admin/items?"

    df = pd.DataFrame(
        columns=[
            "Item Id",
            "Item Name",
            "Type",
            "Description",
            "State",
            "Last Updated Date",
            "Creator Principal Id",
            "Creator Principal Display Name",
            "Creator Principal Type",
            "Creator User Principal Name",
            "Workspace Id",
            "Capacity Id",
        ]
    )

    if workspace is not None:
        workspace = fabric.resolve_workspace_name(workspace)
        workspace_id = fabric.resolve_workspace_id(workspace)
        url += f"workspaceId={workspace_id}&"
    if capacity_name is not None:
        dfC = list_capacities()
        dfC_filt = dfC[dfC["Capacity Name"] == capacity_name]
        if len(dfC_filt) == 0:
            raise ValueError(
                f"{icons.red_dot} The '{capacity_name}' capacity does not exist."
            )
        capacity_id = dfC_filt["Capacity Id"].iloc[0]
        url += f"capacityId={capacity_id}&"
    if state is not None:
        url += f"state={state}&"
    if type is not None:
        url += f"type={type}&"

    if url.endswith("?") or url.endswith("&"):
        url = url[:-1]

    client = fabric.FabricRestClient()
    response = client.get(url)

    if response.status_code != 200:
        raise FabricHTTPException(response)

    responses = pagination(client, response)

    for r in responses:
        for v in r.get("itemEntities", []):
            new_data = {
                "Item Id": v.get("id"),
                "Type": v.get("type"),
                "Item Name": v.get("name"),
                "Description": v.get("description"),
                "State": v.get("state"),
                "Last Updated Date": v.get("lastUpdatedDate"),
                "Creator Principal Id": v.get("creatorPrincipal", {}).get("id"),
                "Creator Principal Display Name": v.get("creatorPrincipal", {}).get(
                    "displayName"
                ),
                "Creator Principal Type": v.get("creatorPrincipal", {}).get("type"),
                "Creator User Principal Name": v.get("creatorPrincipal", {})
                .get("userDetails", {})
                .get("userPrincipalName"),
                "Workspace Id": v.get("workspaceId"),
                "Capacity Id": v.get("capacityId"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def list_activity_events(
    start_time: str,
    end_time: str,
    activity_filter: Optional[str] = None,
    user_id_filter: Optional[str] = None,
):
    """
    Shows a list of audit activity events for a tenant.

    Parameters
    ----------
    start_time : str
        Start date and time of the window for audit event results. Example: "2024-09-25T07:55:00".
    end_time : str
        End date and time of the window for audit event results. Example: "2024-09-25T08:55:00".
    activity_filter : str, default=None
        Filter value for activities. Example: 'viewreport'.
    user_id_filter : str, default=None
        Email address of the user.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of audit activity events for a tenant.
    """

    # https://learn.microsoft.com/en-us/rest/api/power-bi/admin/get-activity-events

    df = pd.DataFrame(
        columns=[
            "Id",
            "Record Type",
            "Creation Time",
            "Operation",
            "Organization Id",
            "User Type",
            "User Key",
            "Workload",
            "Result Status",
            "User Id",
            "Client IP",
            "User Agent",
            "Activity",
            "Workspace Name",
            "Workspace Id",
            "Object Id",
            "Request Id",
            "Object Type",
            "Object Display Name",
            "Experience",
            "Refresh Enforcement Policy",
        ]
    )

    tic = "%27"
    space = "%20"
    client = fabric.PowerBIRestClient()
    base_url = "/v1.0/myorg/admin/activityevents"
    conditions = []

    if activity_filter is not None:
        conditions.append(f"Activity{space}eq{space}{tic}{activity_filter}{tic}")
    if user_id_filter is not None:
        conditions.append(f"UserId{space}eq{space}{tic}{user_id_filter}{tic}")

    filter_value = (
        f"&filter={f'{space}and{space}'.join(conditions)}" if conditions else ""
    )

    full_url = f"{base_url}?startDateTime={tic}{start_time}{tic}&endDateTime={tic}{end_time}{tic}{filter_value}"
    response = client.get(full_url)
    if response.status_code != 200:
        raise FabricHTTPException(response)

    responses = pagination(client, response)

    for r in responses:
        for i in r.get("activityEventEntities", []):
            new_data = {
                "Id": i.get("id"),
                "Record Type": i.get("RecordType"),
                "Creation Time": i.get("CreationTime"),
                "Operation": i.get("Operation"),
                "Organization Id": i.get("OrganizationId"),
                "User Type": i.get("UserType"),
                "User Key": i.get("UserKey"),
                "Workload": i.get("Workload"),
                "Result Status": i.get("ResultStatus"),
                "User Id": i.get("UserId"),
                "Client IP": i.get("ClientIP"),
                "User Agent": i.get("UserAgent"),
                "Activity": i.get("Activity"),
                "Workspace Name": i.get("WorkSpaceName"),
                "Workspace Id": i.get("WorkspaceId"),
                "Object Id": i.get("ObjectId"),
                "Request Id": i.get("RequestId"),
                "Object Type": i.get("ObjectType"),
                "Object Display Name": i.get("ObjectDisplayName"),
                "Experience": i.get("Experience"),
                "Refresh Enforcement Policy": i.get("RefreshEnforcementPolicy"),
            }
            df = pd.concat(
                [df, pd.DataFrame(new_data, index=[0])],
                ignore_index=True,
            )

    return df



import sempy.fabric as fabric
import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    pagination,
)
from sempy.fabric.exceptions import FabricHTTPException


def list_mirrored_warehouses(workspace: Optional[str] = None) -> pd.DataFrame:
    """
    Shows the mirrored warehouses within a workspace.

    Parameters
    ----------
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the mirrored warehouses within a workspace.
    """

    df = pd.DataFrame(
        columns=["Mirrored Warehouse Name", "Mirrored Warehouse Id", "Description"]
    )

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    client = fabric.FabricRestClient()
    response = client.get(f"/v1/workspaces/{workspace_id}/mirroredWarehouses")
    if response.status_code != 200:
        raise FabricHTTPException(response)
    responses = pagination(client, response)

    for r in responses:
        for v in r.get("value", []):

            new_data = {
                "Mirrored Warehouse Name": v.get("displayName"),
                "Mirrored Warehouse Id": v.get("id"),
                "Description": v.get("description"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df



import sempy.fabric as fabric
from typing import Optional, List
import sempy_labs._icons as icons
from sempy_labs._helper_functions import lro
from sempy.fabric.exceptions import FabricHTTPException
import pandas as pd
from uuid import UUID


def resolve_domain_id(domain_name: str) -> UUID:
    """
    Obtains the domain Id for a given domain name.

    Parameters
    ----------
    domain_name : str
        The domain name

    Returns
    -------
    UUID
        The domain Id.
    """

    dfL = list_domains()
    dfL_filt = dfL[dfL["Domain Name"] == domain_name]
    if len(dfL_filt) == 0:
        raise ValueError(f"{icons.red_dot} '{domain_name}' is not a valid domain name.")

    return dfL_filt["Domain ID"].iloc[0]


def list_domains(non_empty_only: bool = False) -> pd.DataFrame:
    """
    Shows a list of domains.

    Parameters
    ----------
    non_empty_only : bool, default=False
        When True, only return domains that have at least one workspace containing an item.
        Defaults to False.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the domains.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/domains/list-domains?tabs=HTTP

    df = pd.DataFrame(
        columns=[
            "Domain ID",
            "Domain Name",
            "Description",
            "Parent Domain ID",
            "Contributors Scope",
        ]
    )

    client = fabric.FabricRestClient()
    url = "/v1/admin/domains"
    if non_empty_only:
        url = f"{url}?nonEmptyOnly=True"
    response = client.get(url)

    if response.status_code != 200:
        raise FabricHTTPException(response)

    for v in response.json().get("domains", []):
        new_data = {
            "Domain ID": v.get("id"),
            "Domain Name": v.get("displayName"),
            "Description": v.get("description"),
            "Parent Domain ID": v.get("parentDomainId"),
            "Contributors Scope": v.get("contributorsScope"),
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def list_domain_workspaces(domain_name: str) -> pd.DataFrame:
    """
    Shows a list of workspaces within the domain.

    Parameters
    ----------
    domain_name : str
        The domain name.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of workspaces within the domain.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/domains/list-domain-workspaces?tabs=HTTP

    domain_id = resolve_domain_id(domain_name)

    df = pd.DataFrame(columns=["Workspace ID", "Workspace Name"])

    client = fabric.FabricRestClient()
    response = client.get(f"/v1/admin/domains/{domain_id}/workspaces")

    if response.status_code != 200:
        raise FabricHTTPException(response)

    for v in response.json().get("value", []):
        new_data = {
            "Workspace ID": v.get("id"),
            "Workspace Name": v.get("displayName"),
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def create_domain(
    domain_name: str,
    description: Optional[str] = None,
    parent_domain_name: Optional[str] = None,
):
    """
    Creates a new domain.

    Parameters
    ----------
    domain_name : str
        The domain name.
    description : str, default=None
        The domain description.
    parent_domain_name : str, default=None
        The parent domain name.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/domains/create-domain?tabs=HTTP

    if parent_domain_name is not None:
        parent_domain_id = resolve_domain_id(parent_domain_name)

    payload = {}
    payload["displayName"] = domain_name
    if description is not None:
        payload["description"] = description
    if parent_domain_name is not None:
        payload["parentDomainId"] = parent_domain_id

    client = fabric.FabricRestClient()
    response = client.post("/v1/admin/domains", json=payload)

    if response.status_code != 201:
        raise FabricHTTPException(response)

    print(f"{icons.green_dot} The '{domain_name}' domain has been created.")


def delete_domain(domain_name: str):
    """
    Deletes a domain.

    Parameters
    ----------
    domain_name : str
        The domain name.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/domains/delete-domain?tabs=HTTP

    domain_id = resolve_domain_id(domain_name)

    client = fabric.FabricRestClient()
    response = client.delete(f"/v1/admin/domains/{domain_id}")

    if response.status_code != 200:
        raise FabricHTTPException(response)

    print(f"{icons.green_dot} The '{domain_name}' domain has been deleted.")


def update_domain(
    domain_name: str,
    description: Optional[str] = None,
    contributors_scope: Optional[str] = None,
):
    """
    Updates a domain's properties.

    Parameters
    ----------
    domain_name : str
        The domain name.
    description : str, default=None
        The domain description.
    contributors_scope : str, default=None
        The domain `contributor scope <https://learn.microsoft.com/rest/api/fabric/admin/domains/update-domain?tabs=HTTP#contributorsscopetype>`_.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/domains/update-domain?tabs=HTTP

    contributors_scopes = ["AdminsOnly", "AllTenant", "SpecificUsersAndGroups"]

    if contributors_scope not in contributors_scopes:
        raise ValueError(
            f"{icons.red_dot} Invalid contributors scope. Valid options: {contributors_scopes}."
        )

    domain_id = resolve_domain_id(domain_name)

    payload = {}
    payload["displayName"] = domain_name

    if description is not None:
        payload["description"] = description
    if contributors_scope is not None:
        payload["contributorsScope"] = contributors_scope

    client = fabric.FabricRestClient()
    response = client.patch(f"/v1/admin/domains/{domain_id}", json=payload)

    if response != 200:
        raise FabricHTTPException(response)

    print(f"{icons.green_dot} The '{domain_name}' domain has been updated.")


def assign_domain_workspaces_by_capacities(
    domain_name: str, capacity_names: str | List[str]
):
    """
    Assigns all workspaces that reside on the specified capacities to the specified domain.

    Parameters
    ----------
    domain_name : str
        The domain name.
    capacity_names : str | List[str]
        The capacity names.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/domains/assign-domain-workspaces-by-capacities?tabs=HTTP

    from sempy_labs.admin import list_capacities

    domain_id = resolve_domain_id(domain_name)

    if isinstance(capacity_names, str):
        capacity_names = [capacity_names]

    dfC = list_capacities()

    # Check for invalid capacities
    invalid_capacities = [
        name for name in capacity_names if name not in dfC["Display Name"].values
    ]

    if len(invalid_capacities) == 1:
        raise ValueError(
            f"{icons.red_dot} The {invalid_capacities} capacity is not valid."
        )
    elif len(invalid_capacities) > 1:
        raise ValueError(
            f"{icons.red_dot} The {invalid_capacities} capacities are not valid."
        )

    # Get list of capacity Ids for the payload
    dfC_filt = dfC[dfC["Display Name"].isin(capacity_names)]
    capacity_list = list(dfC_filt["Id"].str.upper())

    payload = {"capacitiesIds": capacity_list}

    client = fabric.FabricRestClient()
    response = client.post(
        f"/v1/admin/domains/{domain_id}/assignWorkspacesByCapacities",
        json=payload,
    )

    lro(client, response)

    print(
        f"{icons.green_dot} The workspaces in the {capacity_names} capacities have been assigned to the '{domain_name}' domain."
    )


def assign_domain_workspaces(domain_name: str, workspace_names: str | List[str]):
    """
    Assigns workspaces to the specified domain by workspace.

    Parameters
    ----------
    domain_name : str
        The domain name.
    workspace_names : str | List[str]
        The Fabric workspace(s).
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/domains/assign-domain-workspaces-by-ids?tabs=HTTP

    domain_id = resolve_domain_id(domain_name=domain_name)

    if isinstance(workspace_names, str):
        workspace_names = [workspace_names]

    dfW = fabric.list_workspaces()

    # Check for invalid capacities
    invalid_workspaces = [
        name for name in workspace_names if name not in dfW["Name"].values
    ]

    if len(invalid_workspaces) == 1:
        raise ValueError(
            f"{icons.red_dot} The {invalid_workspaces} workspace is not valid."
        )
    elif len(invalid_workspaces) > 1:
        raise ValueError(
            f"{icons.red_dot} The {invalid_workspaces} workspaces are not valid."
        )

    dfW_filt = dfW[dfW["Name"].isin(workspace_names)]
    workspace_list = list(dfW_filt["Id"])

    payload = {"workspacesIds": workspace_list}

    client = fabric.FabricRestClient()
    response = client.post(
        f"/v1/admin/domains/{domain_id}/assignWorkspaces",
        json=payload,
    )

    lro(client, response)

    print(
        f"{icons.green_dot} The {workspace_names} workspaces have been assigned to the '{domain_name}' domain."
    )


def unassign_all_domain_workspaces(domain_name: str):
    """
    Unassigns all workspaces from the specified domain.

    Parameters
    ----------
    domain_name : str
        The domain name.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/domains/unassign-all-domain-workspaces?tabs=HTTP

    domain_id = resolve_domain_id(domain_name=domain_name)

    client = fabric.FabricRestClient()
    response = client.post(f"/v1/admin/domains/{domain_id}/unassignAllWorkspaces")

    if response.status_code != 200:
        raise FabricHTTPException(response)
    print(
        f"{icons.green_dot} All workspaces assigned to the '{domain_name}' domain have been unassigned."
    )


def unassign_domain_workspaces(domain_name: str, workspace_names: str | List[str]):
    """
    Unassigns workspaces from the specified domain by workspace.

    Parameters
    ----------
    domain_name : str
        The domain name.
    workspace_names : str | List[str]
        The Fabric workspace(s).
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/admin/domains/unassign-domain-workspaces-by-ids?tabs=HTTP

    domain_id = resolve_domain_id(domain_name=domain_name)

    if isinstance(workspace_names, str):
        workspace_names = [workspace_names]

    dfW = fabric.list_workspaces()

    # Check for invalid capacities
    invalid_workspaces = [
        name for name in workspace_names if name not in dfW["Name"].values
    ]

    if len(invalid_workspaces) == 1:
        raise ValueError(
            f"{icons.red_dot} The {invalid_workspaces} workspace is not valid."
        )
    elif len(invalid_workspaces) > 1:
        raise ValueError(
            f"{icons.red_dot} The {invalid_workspaces} workspaces are not valid."
        )

    dfW_filt = dfW[dfW["Name"].isin(workspace_names)]
    workspace_list = list(dfW_filt["Id"])

    payload = {"workspacesIds": workspace_list}
    client = fabric.FabricRestClient()
    response = client.post(
        f"/v1/admin/domains/{domain_id}/unassignWorkspaces", json=payload
    )

    if response.status_code != 200:
        raise FabricHTTPException(response)
    print(
        f"{icons.green_dot} The {workspace_names} workspaces assigned to the '{domain_name}' domain have been unassigned."
    )



import sempy
import sempy.fabric as fabric
from sempy_labs.lakehouse import get_lakehouse_columns
from sempy_labs.directlake._dl_helper import get_direct_lake_source
from sempy_labs.tom import connect_semantic_model
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def direct_lake_schema_sync(
    dataset: str,
    workspace: Optional[str] = None,
    add_to_model: bool = False,
    **kwargs,
):
    """
    Shows/adds columns which exist in the lakehouse but do not exist in the semantic model (only for tables in the semantic model).

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    add_to_model : bool, default=False
        If set to True, columns which exist in the lakehouse but do not exist in the semantic model are added to the semantic model. No new tables are added.
    """

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM

    if "lakehouse" in kwargs:
        print(
            "The 'lakehouse' parameter has been deprecated as it is no longer necessary. Please remove this parameter from the function going forward."
        )
        del kwargs["lakehouse"]
    if "lakehouse_workspace" in kwargs:
        print(
            "The 'lakehouse_workspace' parameter has been deprecated as it is no longer necessary. Please remove this parameter from the function going forward."
        )
        del kwargs["lakehouse_workspace"]

    workspace = fabric.resolve_workspace_name(workspace)

    artifact_type, lakehouse_name, lakehouse_id, lakehouse_workspace_id = (
        get_direct_lake_source(dataset=dataset, workspace=workspace)
    )

    if artifact_type == "Warehouse":
        raise ValueError(
            f"{icons.red_dot} This function is only valid for Direct Lake semantic models which source from Fabric lakehouses (not warehouses)."
        )
    lakehouse_workspace = fabric.resolve_workspace_name(lakehouse_workspace_id)

    lc = get_lakehouse_columns(lakehouse_name, lakehouse_workspace)

    with connect_semantic_model(
        dataset=dataset, readonly=False, workspace=workspace
    ) as tom:

        for i, r in lc.iterrows():
            lakeTName = r["Table Name"]
            lakeCName = r["Column Name"]
            dType = r["Data Type"]

            if any(
                p.Source.EntityName == lakeTName
                for p in tom.all_partitions()
                if p.SourceType == TOM.PartitionSourceType.Entity
            ):
                table_name = next(
                    t.Name
                    for t in tom.model.Tables
                    for p in t.Partitions
                    if p.SourceType == TOM.PartitionSourceType.Entity
                    and p.Source.EntityName == lakeTName
                )

                if not any(
                    c.SourceColumn == lakeCName and c.Parent.Name == table_name
                    for c in tom.all_columns()
                ):
                    print(
                        f"{icons.yellow_dot} The '{lakeCName}' column exists in the '{lakeTName}' lakehouse table but not in the '{dataset}' semantic model within the '{workspace}' workspace."
                    )
                    if add_to_model:
                        dt = icons.data_type_mapping.get(dType)
                        tom.add_data_column(
                            table_name=table_name,
                            column_name=lakeCName,
                            source_column=lakeCName,
                            data_type=dt,
                        )
                        print(
                            f"{icons.green_dot} The '{lakeCName}' column in the '{lakeTName}' lakehouse table was added to the '{dataset}' semantic model within the '{workspace}' workspace."
                        )



import sempy.fabric as fabric
import pandas as pd
import sempy_labs._icons as icons
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    lro,
    pagination,
)
from sempy.fabric.exceptions import FabricHTTPException


def create_environment(
    environment: str, description: Optional[str] = None, workspace: Optional[str] = None
):
    """
    Creates a Fabric environment.

    Parameters
    ----------
    environment: str
        Name of the environment.
    description : str, default=None
        A description of the environment.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    request_body = {"displayName": environment}

    if description:
        request_body["description"] = description

    client = fabric.FabricRestClient()
    response = client.post(
        f"/v1/workspaces/{workspace_id}/environments", json=request_body
    )

    lro(client, response, status_codes=[201, 202])

    print(
        f"{icons.green_dot} The '{environment}' environment has been created within the '{workspace}' workspace."
    )


def list_environments(workspace: Optional[str] = None) -> pd.DataFrame:
    """
    Shows the environments within a workspace.

    Parameters
    ----------
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the environments within a workspace.
    """

    df = pd.DataFrame(columns=["Environment Name", "Environment Id", "Description"])

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    client = fabric.FabricRestClient()
    response = client.get(f"/v1/workspaces/{workspace_id}/environments")
    if response.status_code != 200:
        raise FabricHTTPException(response)

    responses = pagination(client, response)

    for r in responses:
        for v in r.get("value", []):
            new_data = {
                "Environment Name": v.get("displayName"),
                "Environment Id": v.get("id"),
                "Description": v.get("description"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def delete_environment(environment: str, workspace: Optional[str] = None):
    """
    Deletes a Fabric environment.

    Parameters
    ----------
    environment: str
        Name of the environment.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    from sempy_labs._helper_functions import resolve_environment_id

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)
    environment_id = resolve_environment_id(
        environment=environment, workspace=workspace
    )

    client = fabric.FabricRestClient()
    response = client.delete(
        f"/v1/workspaces/{workspace_id}/environments/{environment_id}"
    )

    if response.status_code != 200:
        raise FabricHTTPException(response)

    print(
        f"{icons.green_dot} The '{environment}' environment within the '{workspace}' workspace has been deleted."
    )


def publish_environment(environment: str, workspace: Optional[str] = None):
    """
    Publishes a Fabric environment.

    Parameters
    ----------
    environment: str
        Name of the environment.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    # https://learn.microsoft.com/en-us/rest/api/fabric/environment/spark-libraries/publish-environment?tabs=HTTP

    from sempy_labs._helper_functions import resolve_environment_id

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)
    environment_id = resolve_environment_id(
        environment=environment, workspace=workspace
    )

    client = fabric.FabricRestClient()
    response = client.post(
        f"/v1/workspaces/{workspace_id}/environments/{environment_id}/staging/publish"
    )

    lro(client, response)

    print(
        f"{icons.green_dot} The '{environment}' environment within the '{workspace}' workspace has been published."
    )



import sempy
import sempy.fabric as fabric
import re
from sempy_labs._helper_functions import (
    create_relationship_name,
    retry,
    format_dax_object_name,
)
from sempy_labs.tom import connect_semantic_model
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def migrate_model_objects_to_semantic_model(
    dataset: str,
    new_dataset: str,
    workspace: Optional[str] = None,
    new_dataset_workspace: Optional[str] = None,
):
    """
    Adds the rest of the model objects (besides tables/columns) and their properties to a Direct Lake semantic model based on an import/DirectQuery semantic model.

    Parameters
    ----------
    dataset : str
        Name of the import/DirectQuery semantic model.
    new_dataset : str
        Name of the Direct Lake semantic model.
    workspace : str, default=None
        The Fabric workspace name in which the import/DirectQuery semantic model exists.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    new_dataset_workspace : str
        The Fabric workspace name in which the Direct Lake semantic model will be created.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    sempy.fabric._client._utils._init_analysis_services()
    import Microsoft.AnalysisServices.Tabular as TOM
    import System

    if dataset == new_dataset:
        raise ValueError(
            f"{icons.red_dot} The 'dataset' and 'new_dataset' parameters are both set to '{dataset}'. These parameters must be set to different values."
        )

    workspace = fabric.resolve_workspace_name(workspace)

    if new_dataset_workspace is None:
        new_dataset_workspace = workspace

    dfT = fabric.list_tables(dataset=dataset, workspace=workspace)
    dfC = fabric.list_columns(dataset=dataset, workspace=workspace)
    dfM = fabric.list_measures(dataset=dataset, workspace=workspace)
    dfRole = fabric.get_roles(dataset=dataset, workspace=workspace)
    dfRLS = fabric.get_row_level_security_permissions(
        dataset=dataset, workspace=workspace
    )
    dfCI = fabric.list_calculation_items(dataset=dataset, workspace=workspace)
    dfP = fabric.list_perspectives(dataset=dataset, workspace=workspace)
    dfTranslation = fabric.list_translations(dataset=dataset, workspace=workspace)
    dfH = fabric.list_hierarchies(dataset=dataset, workspace=workspace)
    dfPar = fabric.list_partitions(dataset=dataset, workspace=workspace)

    dfP_cc = dfPar[(dfPar["Source Type"] == "Calculated")]
    dfP_fp = dfP_cc[dfP_cc["Query"].str.contains("NAMEOF")]

    print(f"{icons.in_progress} Updating '{new_dataset}' based on '{dataset}'...")

    @retry(
        sleep_time=1,
        timeout_error_message=f"{icons.red_dot} Function timed out after 1 minute",
    )
    def dyn_connect():
        with connect_semantic_model(
            dataset=new_dataset, readonly=True, workspace=new_dataset_workspace
        ) as tom:

            tom.model

    dyn_connect()

    @retry(
        sleep_time=1,
        timeout_error_message=f"{icons.red_dot} Function timed out after 1 minute",
    )
    def dyn_connect2():
        with connect_semantic_model(
            dataset=dataset, readonly=True, workspace=workspace
        ) as tom:

            tom.model

    dyn_connect2()

    with connect_semantic_model(
        dataset=new_dataset, readonly=False, workspace=new_dataset_workspace
    ) as tom:

        isDirectLake = tom.is_direct_lake()

        print(f"\n{icons.in_progress} Updating table properties...")
        for t in tom.model.Tables:
            t.IsHidden = bool(dfT.loc[dfT["Name"] == t.Name, "Hidden"].iloc[0])
            t.Description = dfT.loc[dfT["Name"] == t.Name, "Description"].iloc[0]
            t.DataCategory = dfT.loc[dfT["Name"] == t.Name, "Data Category"].iloc[0]

            print(
                f"{icons.green_dot} The '{t.Name}' table's properties have been updated."
            )

        print(f"\n{icons.in_progress} Updating column properties...")
        for t in tom.model.Tables:
            if (
                t.Name not in dfP_fp["Table Name"].values
            ):  # do not include field parameters
                dfT_filtered = dfT[dfT["Name"] == t.Name]
                tType = dfT_filtered["Type"].iloc[0]
                for c in t.Columns:
                    if not c.Name.startswith("RowNumber-"):
                        dfC_filt = dfC[
                            (dfC["Table Name"] == t.Name)
                            & (dfC["Column Name"] == c.Name)
                        ]
                        cName = dfC_filt["Column Name"].iloc[0]
                        c.Name = cName
                        # if tType == "Table":
                        #    c.SourceColumn = cName.replace(" ", "_")
                        c.IsHidden = bool(dfC_filt["Hidden"].iloc[0])
                        c.DataType = System.Enum.Parse(
                            TOM.DataType, dfC_filt["Data Type"].iloc[0]
                        )
                        c.DisplayFolder = dfC_filt["Display Folder"].iloc[0]
                        c.FormatString = dfC_filt["Format String"].iloc[0]
                        c.SummarizeBy = System.Enum.Parse(
                            TOM.AggregateFunction,
                            dfC_filt["Summarize By"].iloc[0],
                        )
                        c.DataCategory = dfC_filt["Data Category"].iloc[0]
                        c.IsKey = bool(dfC_filt["Key"].iloc[0])
                        sbc = dfC_filt["Sort By Column"].iloc[0]

                        if sbc is not None:
                            if any(
                                o.Name == sbc and o.Parent.Name == c.Parent.Name
                                for o in tom.all_columns()
                            ):
                                c.SortByColumn = tom.model.Tables[t.Name].Columns[sbc]
                            else:
                                print(
                                    f"{icons.red_dot} Failed to create '{sbc}' as a Sort By Column for the '{c.Name}' in the '{t.Name}' table."
                                )
                        print(
                            f"{icons.green_dot} The '{t.Name}'[{c.Name}] column's properties have been updated."
                        )

        print(f"\n{icons.in_progress} Creating hierarchies...")
        dfH_grouped = (
            dfH.groupby(
                [
                    "Table Name",
                    "Hierarchy Name",
                    "Hierarchy Hidden",
                    "Hierarchy Description",
                ]
            )
            .agg({"Level Name": list, "Column Name": list})
            .reset_index()
        )

        for i, r in dfH_grouped.iterrows():
            tName = r["Table Name"]
            hName = r["Hierarchy Name"]
            hDesc = r["Hierarchy Description"]
            hHid = bool(r["Hierarchy Hidden"])
            cols = r["Column Name"]
            lvls = r["Level Name"]

            missing_columns = []
            for col in cols:
                if not any(
                    c.Name == col
                    for t in tom.model.Tables
                    if t.Name == tName
                    for c in t.Columns
                ):
                    missing_columns.append(col)

            if any(
                t.Name == tName and h.Name == hName
                for t in tom.model.Tables
                for h in t.Hierarchies
            ):
                print(
                    f"{icons.warning} The '{hName}' hierarchy within the '{tName}' table already exists."
                )
            elif len(missing_columns) > 0:
                print(
                    f"{icons.red_dot} The '{hName}' hierarchy within the '{tName}' table cannot be created as the {missing_columns} column)s) do not exist."
                )
            else:
                tom.add_hierarchy(
                    table_name=tName,
                    hierarchy_name=hName,
                    hierarchy_description=hDesc,
                    hierarchy_hidden=hHid,
                    columns=cols,
                    levels=lvls,
                )
                print(f"{icons.green_dot} The '{hName}' hierarchy has been added.")

        print(f"\n{icons.in_progress} Creating measures...")
        for i, r in dfM.iterrows():
            tName = r["Table Name"]
            mName = r["Measure Name"]
            mExpr = r["Measure Expression"]
            mHidden = bool(r["Measure Hidden"])
            mDF = r["Measure Display Folder"]
            mDesc = r["Measure Description"]
            mFS = r["Format String"]

            if not any(m.Name == mName for m in tom.all_measures()):
                tom.add_measure(
                    table_name=tName,
                    measure_name=mName,
                    expression=mExpr,
                    hidden=mHidden,
                    display_folder=mDF,
                    description=mDesc,
                    format_string=mFS,
                )
                print(f"{icons.green_dot} The '{mName}' measure has been added.")
        print(f"\n{icons.in_progress} Creating calculation groups...")
        for cgName in dfCI["Calculation Group Name"].unique():

            isHidden = bool(
                dfCI.loc[(dfCI["Calculation Group Name"] == cgName), "Hidden"].iloc[0]
            )
            prec = int(
                dfCI.loc[(dfCI["Calculation Group Name"] == cgName), "Precedence"].iloc[
                    0
                ]
            )
            desc = dfCI.loc[
                (dfCI["Calculation Group Name"] == cgName), "Description"
            ].iloc[0]

            if not any(t.Name == cgName for t in tom.model.Tables):
                tom.add_calculation_group(
                    name=cgName,
                    description=desc,
                    precedence=prec,
                    hidden=isHidden,
                )
                print(
                    f"{icons.green_dot} The '{cgName}' calculation group has been added."
                )
                tom.model.DiscourageImplicitMeasures = True

                # print(
                #    f"\n{icons.in_progress} Updating calculation group column names..."
                # )
                dfC_filt = dfC[(dfC["Table Name"] == cgName) & (dfC["Hidden"] == False)]
                colName = dfC_filt["Column Name"].iloc[0]
                tom.model.Tables[cgName].Columns["Name"].Name = colName

            calcItems = dfCI.loc[
                dfCI["Calculation Group Name"] == cgName,
                "Calculation Item Name",
            ].unique()

            print(f"\n{icons.in_progress} Creating calculation items...")
            for calcItem in calcItems:
                ordinal = int(
                    dfCI.loc[
                        (dfCI["Calculation Group Name"] == cgName)
                        & (dfCI["Calculation Item Name"] == calcItem),
                        "Ordinal",
                    ].iloc[0]
                )
                expr = dfCI.loc[
                    (dfCI["Calculation Group Name"] == cgName)
                    & (dfCI["Calculation Item Name"] == calcItem),
                    "Expression",
                ].iloc[0]
                fse = dfCI.loc[
                    (dfCI["Calculation Group Name"] == cgName)
                    & (dfCI["Calculation Item Name"] == calcItem),
                    "Format String Expression",
                ].iloc[0]

                if not any(
                    ci.CalculationGroup.Parent.Name == cgName and ci.Name == calcItem
                    for ci in tom.all_calculation_items()
                ):
                    tom.add_calculation_item(
                        table_name=cgName,
                        calculation_item_name=calcItem,
                        expression=expr,
                        format_string_expression=fse,
                        ordinal=ordinal,
                    )
                    print(
                        f"{icons.green_dot} The '{calcItem}' has been added to the '{cgName}' calculation group."
                    )

        print(f"\n{icons.in_progress} Creating relationships...")
        with connect_semantic_model(
            dataset=dataset, readonly=True, workspace=workspace
        ) as tom_old:

            for r in tom_old.model.Relationships:
                relName = create_relationship_name(
                    r.FromTable.Name, r.FromColumn.Name, r.ToTable.Name, r.ToColumn.Name
                )

                # Relationship already exists
                if any(
                    rel.FromTable.Name == r.FromTable.Name
                    and rel.FromColumn.Name == r.FromColumn.Name
                    and rel.ToTable.Name == r.ToTable.Name
                    and rel.ToColumn.Name == r.ToColumn.Name
                    for rel in tom.model.Relationships
                ):
                    print(
                        f"{icons.warning} The {relName} relationship was not created as it already exists in the '{new_dataset}' semantic model within the '{new_dataset_workspace}' workspace."
                    )

                # Direct Lake with incompatible column data types
                elif isDirectLake and r.FromColumn.DataType != r.ToColumn.DataType:
                    print(
                        f"{icons.warning} The {relName} relationship was not created as Direct Lake does not support relationships based on columns with different data types."
                    )
                # Direct Lake using DateTime columns
                elif isDirectLake and (
                    r.FromColumn.DataType == TOM.DataType.DateTime
                    or r.ToColumn.DataType == TOM.DataType.DateTime
                ):
                    print(
                        f"{icons.red_dot} The {relName} relationship was not created as Direct Lake does not support relationships based on columns of DateTime data type."
                    )
                # Columns do not exist in the new semantic model
                elif not any(
                    c.Name == r.FromColumn.Name and c.Parent.Name == r.FromTable.Name
                    for c in tom.all_columns()
                ) or not any(
                    c.Name == r.ToColumn.Name and c.Parent.Name == r.ToTable.Name
                    for c in tom.all_columns()
                ):
                    # Direct lake and based on calculated column
                    if isDirectLake and (
                        any(
                            c.Name == r.FromColumn.Name
                            and c.Parent.Name == r.FromTable.Name
                            for c in tom_old.all_calculated_columns()
                        )
                        or any(
                            c.Name == r.ToColumn.Name
                            and c.Parent.Name == r.ToTable.Name
                            for c in tom_old.all_calculated_columns()
                        )
                    ):
                        print(
                            f"{icons.red_dot} The {relName} relationship was not created as the necssary column(s) do not exist. This is due to Direct Lake not supporting calculated columns."
                        )
                    elif not any(
                        c.Name == r.FromColumn.Name
                        and c.Parent.Name == r.FromTable.Name
                        for c in tom.all_columns()
                    ):
                        print(
                            f"{icons.red_dot} The {relName} relationship cannot be created because the {format_dax_object_name(r.FromTable.Name, r.FromColumn.Name)} column is not available in the '{new_dataset}' semantic model within the '{new_dataset_workspace}' workspace."
                        )
                    elif not any(
                        c.Name == r.ToColumn.Name and c.Parent.Name == r.ToTable.Name
                        for c in tom.all_columns()
                    ):
                        print(
                            f"{icons.red_dot} The {relName} relationship cannot be created because the {format_dax_object_name(r.ToTable.Name, r.ToColumn.Name)} column is not available in the '{new_dataset}' semantic model within the '{new_dataset_workspace}' workspace."
                        )
                else:
                    tom.add_relationship(
                        from_table=r.FromTable.Name,
                        from_column=r.FromColumn.Name,
                        to_table=r.ToTable.Name,
                        to_column=r.ToColumn.Name,
                        from_cardinality=str(r.FromCardinality),
                        to_cardinality=str(r.ToCardinality),
                        cross_filtering_behavior=str(r.CrossFilteringBehavior),
                        security_filtering_behavior=str(r.SecurityFilteringBehavior),
                        rely_on_referential_integrity=r.RelyOnReferentialIntegrity,
                        is_active=r.IsActive,
                    )
                    print(
                        f"{icons.green_dot} The {relName} relationship has been added."
                    )

        print(f"\n{icons.in_progress} Creating roles...")
        for index, row in dfRole.iterrows():
            roleName = row["Role"]
            roleDesc = row["Description"]
            modPerm = row["Model Permission"]

            if not any(r.Name == roleName for r in tom.model.Roles):
                tom.add_role(
                    role_name=roleName,
                    model_permission=modPerm,
                    description=roleDesc,
                )
                print(f"{icons.green_dot} The '{roleName}' role has been added.")

        print(f"\n{icons.in_progress} Creating row level security...")
        for index, row in dfRLS.iterrows():
            roleName = row["Role"]
            tName = row["Table"]
            expr = row["Filter Expression"]

            if any(t.Name == tName for t in tom.model.Tables):
                tom.set_rls(
                    role_name=roleName, table_name=tName, filter_expression=expr
                )
                print(
                    f"{icons.green_dot} Row level security for the '{tName}' table within the '{roleName}' role has been set."
                )
            else:
                print(
                    f"{icons.red_dot} Row level security for the '{tName}' table within the '{roleName}' role was not set."
                )

        print(f"\n{icons.in_progress} Creating perspectives...")
        for pName in dfP["Perspective Name"].unique():

            if not any(p.Name == pName for p in tom.model.Perspectives):
                tom.add_perspective(perspective_name=pName)
                print(f"{icons.green_dot} The '{pName}' perspective has been added.")

        print(f"\n{icons.in_progress} Adding objects to perspectives...")
        for index, row in dfP.iterrows():
            pName = row["Perspective Name"]
            tName = row["Table Name"]
            oName = row["Object Name"]
            oType = row["Object Type"]
            tType = dfT.loc[(dfT["Name"] == tName), "Type"].iloc[0]

            try:
                if oType == "Table":
                    tom.add_to_perspective(
                        object=tom.model.Tables[tName], perspective_name=pName
                    )
                elif oType == "Column":
                    tom.add_to_perspective(
                        object=tom.model.Tables[tName].Columns[oName],
                        perspective_name=pName,
                    )
                elif oType == "Measure":
                    tom.add_to_perspective(
                        object=tom.model.Tables[tName].Measures[oName],
                        perspective_name=pName,
                    )
                elif oType == "Hierarchy":
                    tom.add_to_perspective(
                        object=tom.model.Tables[tName].Hierarchies[oName],
                        perspective_name=pName,
                    )
            except Exception:
                pass

        print(f"\n{icons.in_progress} Creating translation languages...")
        for trName in dfTranslation["Culture Name"].unique():
            if not any(c.Name == trName for c in tom.model.Cultures):
                tom.add_translation(trName)
                print(
                    f"{icons.green_dot} The '{trName}' translation language has been added."
                )

        print(f"\n{icons.in_progress} Creating translation values...")
        for index, row in dfTranslation.iterrows():
            trName = row["Culture Name"]
            tName = row["Table Name"]
            oName = row["Object Name"]
            oType = row["Object Type"]
            translation = row["Translation"]
            prop = row["Property"]

            if prop == "Caption":
                prop = "Name"
            elif prop == "DisplayFolder":
                prop = "Display Folder"

            try:
                if oType == "Table":
                    tom.set_translation(
                        object=tom.model.Tables[tName],
                        language=trName,
                        property=prop,
                        value=translation,
                    )
                elif oType == "Column":
                    tom.set_translation(
                        object=tom.model.Tables[tName].Columns[oName],
                        language=trName,
                        property=prop,
                        value=translation,
                    )
                elif oType == "Measure":
                    tom.set_translation(
                        object=tom.model.Tables[tName].Measures[oName],
                        language=trName,
                        property=prop,
                        value=translation,
                    )
                elif oType == "Hierarchy":
                    tom.set_translation(
                        object=tom.model.Tables[tName].Hierarchies[oName],
                        language=trName,
                        property=prop,
                        value=translation,
                    )
                elif oType == "Level":

                    pattern = r"\[([^]]+)\]"
                    matches = re.findall(pattern, oName)
                    lName = matches[0]

                    pattern = r"'([^']+)'"
                    matches = re.findall(pattern, oName)
                    hName = matches[0]
                    tom.set_translation(
                        object=tom.model.Tables[tName].Hierarchies[hName].Levels[lName],
                        language=trName,
                        property=prop,
                        value=translation,
                    )
            except Exception:
                pass

        print(
            f"\n{icons.green_dot} Migration of objects from '{dataset}' -> '{new_dataset}' is complete."
        )



import sempy.fabric as fabric
import pandas as pd
import sempy_labs._icons as icons
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    lro,
    pagination,
)
from sempy.fabric.exceptions import FabricHTTPException


def list_ml_models(workspace: Optional[str] = None) -> pd.DataFrame:
    """
    Shows the ML models within a workspace.

    Parameters
    ----------
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the ML models within a workspace.
    """

    df = pd.DataFrame(columns=["ML Model Name", "ML Model Id", "Description"])

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    client = fabric.FabricRestClient()
    response = client.get(f"/v1/workspaces/{workspace_id}/mlModels")
    if response.status_code != 200:
        raise FabricHTTPException(response)

    responses = pagination(client, response)

    for r in responses:
        for v in r.get("value", []):
            model_id = v.get("id")
            modelName = v.get("displayName")
            desc = v.get("description")

            new_data = {
                "ML Model Name": modelName,
                "ML Model Id": model_id,
                "Description": desc,
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def create_ml_model(
    name: str, description: Optional[str] = None, workspace: Optional[str] = None
):
    """
    Creates a Fabric ML model.

    Parameters
    ----------
    name: str
        Name of the ML model.
    description : str, default=None
        A description of the environment.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    request_body = {"displayName": name}

    if description:
        request_body["description"] = description

    client = fabric.FabricRestClient()
    response = client.post(f"/v1/workspaces/{workspace_id}/mlModels", json=request_body)

    lro(client, response, status_codes=[201, 202])

    print(
        f"{icons.green_dot} The '{name}' ML model has been created within the '{workspace}' workspace."
    )


def delete_ml_model(name: str, workspace: Optional[str] = None):
    """
    Deletes a Fabric ML model.

    Parameters
    ----------
    name: str
        Name of the ML model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace, workspace_id) = resolve_workspace_name_and_id(workspace)

    item_id = fabric.resolve_item_id(
        item_name=name, type="MLModel", workspace=workspace
    )

    client = fabric.FabricRestClient()
    response = client.delete(f"/v1/workspaces/{workspace_id}/mlModels/{item_id}")

    if response.status_code != 200:
        raise FabricHTTPException(response)

    print(
        f"{icons.green_dot} The '{name}' ML model within the '{workspace}' workspace has been deleted."
    )



from sempy_labs.lakehouse._get_lakehouse_columns import get_lakehouse_columns
from sempy_labs.lakehouse._get_lakehouse_tables import get_lakehouse_tables
from sempy_labs.lakehouse._lakehouse import (
    lakehouse_attached,
    optimize_lakehouse_tables,
    vacuum_lakehouse_tables,
)

from sempy_labs.lakehouse._shortcuts import (
    # create_shortcut,
    create_shortcut_onelake,
    delete_shortcut,
)

__all__ = [
    "get_lakehouse_columns",
    "get_lakehouse_tables",
    "lakehouse_attached",
    "optimize_lakehouse_tables",
    # create_shortcut,
    "create_shortcut_onelake",
    "delete_shortcut",
    "vacuum_lakehouse_tables",
]



import sempy.fabric as fabric
import pandas as pd
from typing import Optional, Union, List
from sempy._utils._log import log
import struct
import uuid
from itertools import chain, repeat
from sempy.fabric.exceptions import FabricHTTPException
from sempy_labs._helper_functions import resolve_warehouse_id, resolve_lakehouse_id


def _bytes2mswin_bstr(value: bytes) -> bytes:
    """Convert a sequence of bytes into a (MS-Windows) BSTR (as bytes).

    See https://github.com/mkleehammer/pyodbc/issues/228#issuecomment-319190980
    for the original code.  It appears the input is converted to an
    MS-Windows BSTR (in 'Little-endian' format).

    See https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-dtyp\
       /692a42a9-06ce-4394-b9bc-5d2a50440168
    for more info on BSTR.

    :param value: the sequence of bytes to convert
    :return: the converted value (as a sequence of bytes)
    """

    encoded_bytes = bytes(chain.from_iterable(zip(value, repeat(0))))
    return struct.pack("<i", len(encoded_bytes)) + encoded_bytes


class ConnectBase:
    def __init__(
        self,
        name: str,
        workspace: Optional[Union[str, uuid.UUID]] = None,
        timeout: Optional[int] = None,
        endpoint_type: str = "warehouse",
    ):
        from sempy.fabric._token_provider import SynapseTokenProvider
        import pyodbc

        workspace = fabric.resolve_workspace_name(workspace)
        workspace_id = fabric.resolve_workspace_id(workspace)

        # Resolve the appropriate ID (warehouse or lakehouse)
        if endpoint_type == "warehouse":
            resource_id = resolve_warehouse_id(warehouse=name, workspace=workspace)
        else:
            resource_id = resolve_lakehouse_id(lakehouse=name, workspace=workspace)

        # Get the TDS endpoint
        client = fabric.FabricRestClient()
        response = client.get(
            f"v1/workspaces/{workspace_id}/{endpoint_type}s/{resource_id}"
        )
        if response.status_code != 200:
            raise FabricHTTPException(response)

        if endpoint_type == "warehouse":
            tds_endpoint = response.json().get("properties", {}).get("connectionString")
        else:
            tds_endpoint = (
                response.json()
                .get("properties", {})
                .get("sqlEndpointProperties", {})
                .get("connectionString")
            )

        # Set up the connection string
        access_token = SynapseTokenProvider()()
        tokenstruct = _bytes2mswin_bstr(access_token.encode())
        conn_str = f"DRIVER={{ODBC Driver 18 for SQL Server}};SERVER={tds_endpoint};DATABASE={name};Encrypt=Yes;"

        if timeout is not None:
            conn_str += f"Connect Timeout={timeout};"

        self.connection = pyodbc.connect(conn_str, attrs_before={1256: tokenstruct})

    @log
    def query(
        self, sql: Union[str, List[str]]
    ) -> Union[List[pd.DataFrame], pd.DataFrame, None]:
        """
        Runs a SQL or T-SQL query (or multiple queries) against a Fabric Warehouse/Lakehouse.

        Parameters
        ----------
        sql : str or List[str]
            A single SQL or T-SQL query, or a list of queries to be executed.

        Returns
        -------
        Union[List[pandas.DataFrame], pandas.DataFrame, None]
            A list of pandas DataFrames if multiple SQL queries return results,
            a single DataFrame if one query is executed and returns results, or None.
        """

        cursor = None
        results = []

        if isinstance(sql, str):
            sql = [sql]

        try:
            cursor = self.connection.cursor()

            for sql_query in sql:
                cursor.execute(sql_query)

                # Commit for non-select queries (like CREATE, INSERT, etc.)
                if not cursor.description:
                    self.connection.commit()
                else:
                    # Fetch and append results for queries that return a result set
                    result = pd.DataFrame.from_records(
                        cursor.fetchall(),
                        columns=[col[0] for col in cursor.description],
                    )
                    results.append(result)

            # Return results if any queries returned a result set
            return results if len(results) > 1 else (results[0] if results else None)

        finally:
            if cursor:
                cursor.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        self.connection.close()


class ConnectWarehouse(ConnectBase):
    def __init__(
        self,
        warehouse: str,
        workspace: Optional[Union[str, uuid.UUID]] = None,
        timeout: Optional[int] = None,
    ):
        super().__init__(
            name=warehouse,
            workspace=workspace,
            timeout=timeout,
            endpoint_type="warehouse",
        )


class ConnectLakehouse(ConnectBase):
    def __init__(
        self,
        lakehouse: str,
        workspace: Optional[Union[str, uuid.UUID]] = None,
        timeout: Optional[int] = None,
    ):
        super().__init__(
            name=lakehouse,
            workspace=workspace,
            timeout=timeout,
            endpoint_type="lakehouse",
        )



import sempy.fabric as fabric
import pandas as pd
from sempy_labs._helper_functions import format_dax_object_name
import sempy_labs._icons as icons
from typing import Any, Dict, Optional
from anytree import Node, RenderTree
from sempy._utils._log import log


@log
def get_measure_dependencies(
    dataset: str, workspace: Optional[str] = None
) -> pd.DataFrame:
    """
    Shows all dependencies for all measures in a semantic model.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        Shows all dependencies for all measures in the semantic model.
    """

    workspace = fabric.resolve_workspace_name(workspace)

    dep = fabric.evaluate_dax(
        dataset=dataset,
        workspace=workspace,
        dax_string="""
        SELECT
         [TABLE] AS [Table Name]
        ,[OBJECT] AS [Object Name]
        ,[OBJECT_TYPE] AS [Object Type]
        ,[REFERENCED_TABLE] AS [Referenced Table]
        ,[REFERENCED_OBJECT] AS [Referenced Object]
        ,[REFERENCED_OBJECT_TYPE] AS [Referenced Object Type]
        FROM $SYSTEM.DISCOVER_CALC_DEPENDENCY
        WHERE [OBJECT_TYPE] = 'MEASURE'
        """,
    )

    dep["Object Type"] = dep["Object Type"].str.capitalize()
    dep["Referenced Object Type"] = dep["Referenced Object Type"].str.capitalize()

    dep["Full Object Name"] = format_dax_object_name(
        dep["Table Name"], dep["Object Name"]
    )
    dep["Referenced Full Object Name"] = format_dax_object_name(
        dep["Referenced Table"], dep["Referenced Object"]
    )
    dep["Parent Node"] = dep["Object Name"]

    df = dep

    df["Done"] = df.apply(
        lambda row: False if row["Referenced Object Type"] == "Measure" else True,
        axis=1,
    )

    while any(df["Done"] == False):
        for i, r in df.iterrows():
            rObjFull = r["Referenced Full Object Name"]
            rObj = r["Referenced Object"]
            if r["Done"] == False:
                dep_filt = dep[dep["Full Object Name"] == rObjFull]

                for index, dependency in dep_filt.iterrows():
                    d = True
                    if dependency.iloc[5] == "Measure":
                        d = False
                        df = pd.concat(
                            [
                                df,
                                pd.DataFrame(
                                    [
                                        {
                                            "Table Name": r["Table Name"],
                                            "Object Name": r["Object Name"],
                                            "Object Type": r["Object Type"],
                                            "Referenced Object": dependency.iloc[4],
                                            "Referenced Table": dependency.iloc[3],
                                            "Referenced Object Type": dependency.iloc[
                                                5
                                            ],
                                            "Done": d,
                                            "Full Object Name": r["Full Object Name"],
                                            "Referenced Full Object Name": dependency.iloc[
                                                7
                                            ],
                                            "Parent Node": rObj,
                                        }
                                    ]
                                ),
                            ],
                            ignore_index=True,
                        )
                    else:
                        df = pd.concat(
                            [
                                df,
                                pd.DataFrame(
                                    [
                                        {
                                            "Table Name": r["Table Name"],
                                            "Object Name": r["Object Name"],
                                            "Object Type": r["Object Type"],
                                            "Referenced Object": dependency.iloc[4],
                                            "Referenced Table": dependency.iloc[3],
                                            "Referenced Object Type": dependency.iloc[
                                                5
                                            ],
                                            "Done": d,
                                            "Full Object Name": r["Full Object Name"],
                                            "Referenced Full Object Name": dependency.iloc[
                                                7
                                            ],
                                            "Parent Node": rObj,
                                        }
                                    ]
                                ),
                            ],
                            ignore_index=True,
                        )

            df.loc[i, "Done"] = True

    df = df.drop(["Done", "Full Object Name", "Referenced Full Object Name"], axis=1)

    return df


@log
def get_model_calc_dependencies(
    dataset: str, workspace: Optional[str] = None
) -> pd.DataFrame:
    """
    Shows all dependencies for all objects in a semantic model.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        Shows all dependencies for all objects in the semantic model.
    """

    workspace = fabric.resolve_workspace_name(workspace)

    dep = fabric.evaluate_dax(
        dataset=dataset,
        workspace=workspace,
        dax_string="""
        SELECT
        [TABLE] AS [Table Name]
        ,[OBJECT] AS [Object Name]
        ,[OBJECT_TYPE] AS [Object Type]
        ,[EXPRESSION] AS [Expression]
        ,[REFERENCED_TABLE] AS [Referenced Table]
        ,[REFERENCED_OBJECT] AS [Referenced Object]
        ,[REFERENCED_OBJECT_TYPE] AS [Referenced Object Type]
        FROM $SYSTEM.DISCOVER_CALC_DEPENDENCY
        """,
    )

    dep["Object Type"] = dep["Object Type"].str.replace("_", " ").str.title()
    dep["Referenced Object Type"] = (
        dep["Referenced Object Type"].str.replace("_", " ").str.title()
    )

    dep["Full Object Name"] = format_dax_object_name(
        dep["Table Name"], dep["Object Name"]
    )
    dep["Referenced Full Object Name"] = format_dax_object_name(
        dep["Referenced Table"], dep["Referenced Object"]
    )
    dep["Parent Node"] = dep["Object Name"]

    df = dep

    objs = ["Measure", "Calc Column", "Calculation Item", "Calc Table"]

    df["Done"] = df.apply(
        lambda row: False if row["Referenced Object Type"] in objs else True, axis=1
    )

    while any(df["Done"] == False):
        for i, r in df.iterrows():
            rObjFull = r["Referenced Full Object Name"]
            rObj = r["Referenced Object"]
            if r["Done"] == False:
                dep_filt = dep[dep["Full Object Name"] == rObjFull]

                for index, dependency in dep_filt.iterrows():
                    d = True
                    if dependency.iloc[5] in objs:
                        d = False
                        df = pd.concat(
                            [
                                df,
                                pd.DataFrame(
                                    [
                                        {
                                            "Table Name": r["Table Name"],
                                            "Object Name": r["Object Name"],
                                            "Object Type": r["Object Type"],
                                            "Referenced Object": dependency.iloc[4],
                                            "Referenced Table": dependency.iloc[3],
                                            "Referenced Object Type": dependency.iloc[
                                                5
                                            ],
                                            "Done": d,
                                            "Full Object Name": r["Full Object Name"],
                                            "Referenced Full Object Name": dependency.iloc[
                                                7
                                            ],
                                            "Parent Node": rObj,
                                        }
                                    ]
                                ),
                            ],
                            ignore_index=True,
                        )
                    else:
                        df = pd.concat(
                            [
                                df,
                                pd.DataFrame(
                                    [
                                        {
                                            "Table Name": r["Table Name"],
                                            "Object Name": r["Object Name"],
                                            "Object Type": r["Object Type"],
                                            "Referenced Object": dependency.iloc[5],
                                            "Referenced Table": dependency.iloc[4],
                                            "Referenced Object Type": dependency.iloc[
                                                6
                                            ],
                                            "Done": d,
                                            "Full Object Name": r["Full Object Name"],
                                            "Referenced Full Object Name": dependency.iloc[
                                                7
                                            ],
                                            "Parent Node": rObj,
                                        }
                                    ]
                                ),
                            ],
                            ignore_index=True,
                        )

            df.loc[i, "Done"] = True

    df = df.drop(["Done"], axis=1)

    return df


@log
def measure_dependency_tree(
    dataset: str, measure_name: str, workspace: Optional[str] = None
):
    """
    Prints a measure dependency tree of all dependent objects for a measure in a semantic model.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    measure_name : str
        Name of the measure.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------

    """

    workspace = fabric.resolve_workspace_name(workspace)

    dfM = fabric.list_measures(dataset=dataset, workspace=workspace)
    dfM_filt = dfM[dfM["Measure Name"] == measure_name]

    if len(dfM_filt) == 0:
        print(
            f"{icons.red_dot} The '{measure_name}' measure does not exist in the '{dataset}' semantic model in the '{workspace}' workspace."
        )
        return

    md = get_measure_dependencies(dataset, workspace)
    df_filt = md[md["Object Name"] == measure_name]

    # Create a dictionary to hold references to nodes
    node_dict: Dict[str, Any] = {}

    # Populate the tree
    for _, row in df_filt.iterrows():
        # measure_name = row['Object Name']
        ref_obj_table_name = row["Referenced Table"]
        ref_obj_name = row["Referenced Object"]
        ref_obj_type = row["Referenced Object Type"]
        parent_node_name = row["Parent Node"]

        # Create or get the parent node
        parent_node = node_dict.get(parent_node_name)
        if parent_node is None:
            parent_node = Node(parent_node_name)
            node_dict[parent_node_name] = parent_node
        parent_node.custom_property = f"{icons.measure_icon} "

        # Create the child node
        child_node_name = ref_obj_name
        child_node = Node(child_node_name, parent=parent_node)
        if ref_obj_type == "Column":
            child_node.custom_property = f"{icons.column_icon} '{ref_obj_table_name}'"
        elif ref_obj_type == "Table":
            child_node.custom_property = f"{icons.table_icon} "
        elif ref_obj_type == "Measure":
            child_node.custom_property = f"{icons.measure_icon} "

        # Update the dictionary with the child node
        node_dict[child_node_name] = child_node

    # Visualize the tree structure using RenderTree
    for pre, _, node in RenderTree(node_dict[measure_name]):
        if icons.table_icon in node.custom_property:
            print(f"{pre}{node.custom_property}'{node.name}'")
        else:
            print(f"{pre}{node.custom_property}[{node.name}]")



import sempy.fabric as fabric
from typing import Optional
import pandas as pd
from sempy_labs._helper_functions import (
    format_dax_object_name,
)
from sempy_labs.report._reportwrapper import ReportWrapper
from sempy_labs._list_functions import list_reports_using_semantic_model


def list_unused_objects_in_reports(
    dataset: str, workspace: Optional[str] = None
) -> pd.DataFrame:
    """
    Shows a list of all columns in the semantic model which are not used in any related Power BI reports (including dependencies).

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of all columns in the semantic model which are not used in any related Power BI reports (including dependencies).
    """

    # TODO: what about relationships/RLS?

    dfR = _list_all_report_semantic_model_objects(dataset=dataset, workspace=workspace)
    dfR_filt = (
        dfR[dfR["Object Type"] == "Column"][["Table Name", "Object Name"]]
        .drop_duplicates()
        .reset_index(drop=True)
    )
    dfR_filt["Column Object"] = format_dax_object_name(
        dfR_filt["Table Name"], dfR_filt["Object Name"]
    )

    dfC = fabric.list_columns(dataset=dataset, workspace=workspace)
    dfC["Column Object"] = format_dax_object_name(dfC["Table Name"], dfC["Column Name"])

    df = dfC[~(dfC["Column Object"].isin(dfR_filt["Column Object"].values))]
    df = df.drop("Column Object", axis=1)

    return df


def _list_all_report_semantic_model_objects(
    dataset: str, workspace: Optional[str] = None
) -> pd.DataFrame:
    """
    Shows a unique list of all semantic model objects (columns, measures, hierarchies) which are used in all reports which leverage the semantic model.

    Parameters
    ----------
    dataset : str
        Name of the semantic model.
    workspace : str, default=None
        The Fabric workspace name.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe.
    """

    dfR = list_reports_using_semantic_model(dataset=dataset, workspace=workspace)
    dfs = []

    for _, r in dfR.iterrows():
        report_name = r["Report Name"]
        report_workspace = r["Report Workspace Name"]

        rpt = ReportWrapper(report=report_name, workspace=report_workspace)

        new_data = rpt._list_all_semantic_model_objects()
        new_data["Report Name"] = report_name
        new_data["Report Workspace"] = report_workspace
        dfs.append(new_data)

    df = pd.concat(dfs, ignore_index=True)

    colName = "Report Name"
    df.insert(2, colName, df.pop(colName))
    colName = "Report Workspace"
    df.insert(3, colName, df.pop(colName))

    return df
